"""
Module implementing an interface with
`Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__.
"""

# ruff: noqa: FA100
# do not use `from __future__ import annotations`, it clashes with `dataclass_json`
import datetime as dt
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, Iterator, Optional, Union

from dataclasses_json import CatchAll, LetterCase, Undefined, dataclass_json
from dataclasses_json import config as dataclass_config
from typing_extensions import Literal

from ...constants import RequestType
from ...types import Json, JsonDict
from ..base import SentinelHubFeatureIterator
from ..process import SentinelHubRequest
from ..utils import AccessSpecification, datetime_config, enum_config, remove_undefined, s3_specification
from .base import BaseBatchClient, BaseBatchRequest, BatchRequestStatus, BatchUserAction, StoppedStatusReason

LOGGER = logging.getLogger(__name__)

BatchRequestType = Union[str, dict, "BatchProcessRequest"]


class BatchProcessClient(BaseBatchClient):
    """An interface class for Sentinel Hub Batch API version 2.

    `Batch Process API <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__
    """

    s3_specification = staticmethod(s3_specification)

    # pylint: disable=too-many-public-methods
    @staticmethod
    def _get_service_url(base_url: str) -> str:
        """Provides URL to Catalog API"""
        return f"{base_url}/api/v2/batch"

    def _get_processing_url(self, request_id: Optional[str] = None) -> str:
        """Creates a URL for process endpoint"""
        url = f"{self.service_url}/process"
        if request_id is None:
            return url
        return f"{url}/{request_id}"

    def create(
        self,
        process_request: Union[SentinelHubRequest, JsonDict],
        input: Dict[str, Any],  # noqa: A002  #pylint: disable=redefined-builtin
        output: Dict[str, Any],
        instance_type: Literal["normal", "large"] = "normal",
        description: Optional[str] = None,
        **kwargs: Any,
    ) -> "BatchProcessRequest":
        """Create a new batch request

        `Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__

        :param process_request: An instance of SentinelHubRequest class containing all request parameters.
            Alternatively, it can also be just a payload dictionary for Process API request
        :param input: A dictionary with input parameters. It can be built with `tiling_grid_input` or `geopackage_input`
            methods.
        :param output: A dictionary with output parameters. It can be built with `raster_output` or `zarr_output`
            methods.
        :param instance_type": Specifies which size of instances to use for the request.
        :param description: A description of a batch request
        :param kwargs: Any other arguments to be added to a dictionary of parameters.
        """

        if isinstance(process_request, SentinelHubRequest):
            request_dict = process_request.download_list[0].post_values
        else:
            request_dict = process_request

        if not isinstance(request_dict, dict):
            raise ValueError(
                "Parameter sentinelhub_request should be an instance of SentinelHubRequest or a "
                "dictionary with a request payload"
            )

        payload = remove_undefined(
            {
                "processRequest": request_dict,
                "input": input,
                "output": output,
                "instanceType": instance_type,
                "description": description,
                **kwargs,
            }
        )

        request_info = self.client.get_json_dict(self._get_processing_url(), post_values=payload, use_session=True)

        return BatchProcessRequest.from_dict(request_info)

    @staticmethod
    def geopackage_input(geopackage_specification: AccessSpecification) -> JsonDict:
        """A helper method to build a suitable dictionary for the `input` field.

        :param geopackage_specification: A specification of the S3 path for the Geopackage. Can be built using the
            `s3_specification` helper method.
        """
        return {"type": "geopackage", "features": geopackage_specification}

    @staticmethod
    def tiling_grid_input(
        grid_id: int, resolution: float, buffer_x: Optional[int] = None, buffer_y: Optional[int] = None, **kwargs: Any
    ) -> JsonDict:
        """A helper method to build a dictionary with tiling grid parameters for the `input` field.

        :param grid_id: An ID of a tiling grid
        :param resolution: A grid resolution
        :param buffer_x: Will expand each output tile horizontally (left and right) by specified number of pixels.
        :param buffer_y: Will expand each output tile vertically (up and down) by specified number of pixels.
        :param kwargs: Any other arguments to be added to a dictionary of parameters
        """
        return remove_undefined(
            {
                "type": "tiling-grid",
                "id": grid_id,
                "resolution": resolution,
                "bufferX": buffer_x,
                "bufferY": buffer_y,
                **kwargs,
            }
        )

    @staticmethod
    def raster_output(
        delivery: AccessSpecification,
        *,
        overwrite: Optional[bool] = None,
        skip_existing: Optional[bool] = None,
        cog_output: Optional[bool] = None,
        cog_parameters: Optional[Dict[str, Any]] = None,
        create_collection: Optional[bool] = None,
        collection_id: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """A helper method to build a dictionary specifying raster output

        :param delivery: An S3 access specification containing a path or a template on an s3 bucket where to store
            results. You can use the `s3_specification` method for construction. For more information on templates see
            documentation.
        :param overwrite: A flag specifying if a request should overwrite existing outputs without failing
        :param skip_existing: A flag specifying if existing outputs should be overwritten
        :param cog_output: A flag specifying if outputs should be written in COGs (cloud-optimized GeoTIFFs) or
            normal GeoTIFFs
        :param cog_parameters: A dictionary specifying COG creation parameters. See documentation for more info.
        :param create_collection: If True the results will be written in COGs and a batch collection will be created
        :param collection_id: If True results will be added to an existing collection
        :param kwargs: Any other arguments to be added to a dictionary of parameters
        """
        return remove_undefined(
            {
                "type": "raster",
                "delivery": delivery,
                "overwrite": overwrite,
                "skipExisting": skip_existing,
                "cogOutput": cog_output,
                "cogParameters": cog_parameters,
                "createCollection": create_collection,
                "collectionId": collection_id,
                **kwargs,
            }
        )

    @staticmethod
    def zarr_output(
        delivery: AccessSpecification,
        *,
        group: Optional[Dict[str, Any]] = None,
        array_parameters: Optional[Dict[str, Any]] = None,
        array_overrides: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> JsonDict:
        """A helper method to build a dictionary specifying Zarr output. See documentation for more information on
        each parameter.

        `Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__

        :param delivery: An S3 access specification containing a path or a template on an s3 bucket where to store
            results. You can use the `s3_specification` method for construction. For more information on templates see
            documentation.
        :param group: Zarr group level parameters
        :param array_parameters: Parameters that will be used for all output arrays, except where overriden with
            `array_overrides`. Required unless `array_overrides` includes all required fields for all output arrays.
        :param array_overrides: Overrides the values of `array_arameters` for individual arrays.
        :param kwargs: Any other arguments to be added to a dictionary of parameters
        """
        return remove_undefined(
            {
                "type": "zarr",
                "delivery": delivery,
                "group": group,
                "arrayParameter": array_parameters,
                "arrayOverrides": array_overrides,
                **kwargs,
            }
        )

    def iter_requests(
        self, user_id: Optional[str] = None, search: Optional[str] = None, sort: Optional[str] = None, **kwargs: Any
    ) -> Iterator["BatchProcessRequest"]:
        """Iterate existing batch requests

        `Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__

        :param user_id: Filter requests by a user id who defined a request
        :param search: A search query to filter requests
        :param sort: A sort query
        :param kwargs: Any additional parameters to include in a request query
        :return: An iterator over existing batch requests
        """
        params = remove_undefined({"userid": user_id, "search": search, "sort": sort, **kwargs})
        feature_iterator = SentinelHubFeatureIterator(
            client=self.client, url=self._get_processing_url(), params=params, exception_message="No requests found"
        )
        for request_info in feature_iterator:
            yield BatchProcessRequest.from_dict(request_info)

    def get_request(self, batch_request: BatchRequestType) -> "BatchProcessRequest":
        """Collects information about a single batch request.

        `Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__
        """
        request_id = self._parse_request_id(batch_request)
        request_info = self.client.get_json_dict(url=self._get_processing_url(request_id), use_session=True)
        return BatchProcessRequest.from_dict(request_info)

    def update_request(self, batch_request: BatchRequestType, description: str) -> Json:
        """Update certain batch job request parameters. Can only update requests that are not currently being processed.

        `Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__

        :param batch_request: Batch request ID, a dictionary containing an "ID" field, or a BatchProcessRequest.
        :param description: A description of a batch request to be updated.
        """
        request_id = self._parse_request_id(batch_request)

        return self.client.get_json(
            url=self._get_processing_url(request_id),
            post_values={"description": description},
            request_type=RequestType.PUT,
            use_session=True,
        )

    def start_analysis(self, batch_request: BatchRequestType) -> Json:
        """Starts analysis of a batch job request

        :param batch_request: Batch request ID, a dictionary containing an "ID" field, or a BatchProcessRequest.
        """
        return self._call_job(batch_request, "analyse")

    def start_job(self, batch_request: BatchRequestType) -> Json:
        """Starts running a batch job

        :param batch_request: Batch request ID, a dictionary containing an "ID" field, or a BatchProcessRequest.
        """
        return self._call_job(batch_request, "start")

    def stop_job(self, batch_request: BatchRequestType) -> Json:
        """Stops a batch job

        :param batch_request: Batch request ID, a dictionary containing an "ID" field, or a BatchProcessRequest.
        """
        return self._call_job(batch_request, "stop")

    def iter_tiling_grids(self, **kwargs: Any) -> SentinelHubFeatureIterator:
        """An iterator over tiling grids

        `Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__

        :param kwargs: Any other request query parameters
        :return: An iterator over tiling grid definitions
        """
        return SentinelHubFeatureIterator(
            client=self.client,
            url=f"{self.service_url}/tilinggrids",
            params=remove_undefined(kwargs),
            exception_message="Failed to obtain information about available tiling grids",
        )

    def get_tiling_grid(self, grid_id: int) -> JsonDict:
        """Provides a single tiling grid

        `Batch Process V2 <https://docs.sentinel-hub.com/api/latest/api/batchv2/>`__

        :param grid_id: An ID of a requested tiling grid
        :return: A tiling grid definition
        """
        url = f"{self.service_url}/tilinggrids/{grid_id}"
        return self.client.get_json_dict(url=url, use_session=True)


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass(repr=False)
class BatchProcessRequest(BaseBatchRequest):  # pylint: disable=abstract-method
    """A dataclass object that holds information about a batch request"""

    # dataclass_json doesn't handle parameter inheritance correctly
    # pylint: disable=invalid-name

    request_id: str = field(metadata=dataclass_config(field_name="id"))
    request: dict
    domain_account_id: str
    status: BatchRequestStatus = field(metadata=enum_config(BatchRequestStatus))
    error: Optional[str] = None
    user_action: Optional[BatchUserAction] = field(metadata=enum_config(BatchUserAction), default=None)
    user_action_updated: Optional[dt.datetime] = field(metadata=datetime_config, default=None)
    created: Optional[dt.datetime] = field(metadata=datetime_config, default=None)
    completion_percentage: float = 0
    last_updated: Optional[dt.datetime] = field(metadata=datetime_config, default=None)
    cost_PU: Optional[float] = field(metadata=dataclass_config(field_name="costPU"), default=None)  # noqa: N815
    stopped_status_reason: Optional[StoppedStatusReason] = field(
        metadata=enum_config(StoppedStatusReason), default=None
    )

    other_data: CatchAll = field(default_factory=dict)

    _REPR_PARAM_NAMES = ("request_id", "created", "status", "completion_percentage", "user_action", "cost_PU")



"""
The part of the package that implements interface with Sentinel Hub services.
"""

from .base import BatchRequestStatus, BatchUserAction
from .process import BatchProcessClient, BatchProcessRequest
from .statistical import BatchStatisticalRequest, SentinelHubBatchStatistical
from .utils import (
    monitor_batch_process_analysis,
    monitor_batch_process_job,
    monitor_batch_statistical_analysis,
    monitor_batch_statistical_job,
)



"""
Module with statistics to dataframe transformation.
"""

from __future__ import annotations

from typing import Any, Iterable

from .time_utils import parse_time
from .types import JsonDict

_PANDAS_IMPORT_MESSAGE = (
    "To use this function you need to install the `pandas` library, which is not a dependency of sentinelhub-py."
)
_FULL_TIME_RANGE = "full time range"


def _extract_hist(hist_data: list[dict[str, float]]) -> tuple[list[float], list[float]]:
    """Transform Statistical API histogram into sequences of bins and counts

    :param hist_data: An input representation of Statistical API histogram data in a form of the low edges,
        the high edges, and the counts for each bin.
    :return: Statistical histogram bins and counts value as sequences.
    """

    nbins = len(hist_data)
    bins, counts = [], []
    for idx in range(nbins):
        bins.append(hist_data[idx]["lowEdge"])
        if idx == nbins - 1:
            bins.append(hist_data[idx]["highEdge"])
        counts.append(hist_data[idx]["count"])

    return bins, counts


def _extract_stats(interval_output: JsonDict, exclude_stats: list[str]) -> dict[str, list[float] | float]:
    """Transform statistics into pandas.DataFrame entry

    :param interval_output: An input representation of statistics of an aggregation interval.
    :param exclude_stats: Statistics that will be excluded from output.
    :return: Statistics as a pandas.DataFrame entry.
    """
    stat_entry: dict[str, list[float] | float] = {}
    for output_name, output_data in interval_output.items():  # pylint: disable=too-many-nested-blocks
        for band_name, band_values in output_data["bands"].items():
            band_stats = band_values["stats"]
            # statistics are not valid when sample count equals to no data count
            if band_stats["sampleCount"] == band_stats["noDataCount"]:
                break

            for stat_name, value in band_stats.items():
                if stat_name not in exclude_stats:
                    col_name = f"{output_name}_{band_name}_{stat_name}"
                    if stat_name == "percentiles":
                        for percentile_name, percentile_value in value.items():
                            stat_entry[f"{col_name}_{percentile_name}"] = percentile_value
                    else:
                        stat_entry[col_name] = value

            if "histogram" in band_values:
                band_bins = band_values["histogram"]["bins"]
                hist_bins, hist_counts = _extract_hist(band_bins)
                stat_entry[f"{output_name}_{band_name}_bins"] = hist_bins
                stat_entry[f"{output_name}_{band_name}_counts"] = hist_counts

    return stat_entry


def _extract_response_data(response_data: list[JsonDict], exclude_stats: list[str]) -> list[JsonDict]:
    """Transform Statistical API response into a pandas.DataFrame

    :param response_data: An input representation of Statistical API response. The response is a list of JsonDict and
        each contains the statistics of an aggregation interval.
    :param exclude_stats: Statistics that will be excluded from output.
    :return: DataFrame entries of all aggregation intervals of a single geometry.
    """
    df_entries = []
    for interval in response_data:
        if "outputs" in interval:
            df_entry: dict[str, Any] = _extract_stats(interval["outputs"], exclude_stats)
            if df_entry:
                df_entry["interval_from"] = parse_time(interval["interval"]["from"])
                df_entry["interval_to"] = parse_time(interval["interval"]["to"])
                df_entries.append(df_entry)

    return df_entries


def _is_batch_stat(result_data: JsonDict) -> bool:
    """Identifies whether the resulting data belongs to a batch statistical request or not"""
    return "id" in result_data


def _is_valid_batch_response(result_data: JsonDict) -> bool:
    """Identifies whether there is a valid batch response"""
    return "error" not in result_data and result_data["response"]["status"] == "OK"


def statistical_to_dataframe(result_data: list[JsonDict], exclude_stats: list[str] | None = None) -> Any:
    """Transform (Batch) Statistical API results into a pandas.DataFrame

    This function has a dependency of the `pandas` library, which is not a requirement of sentinelhub-py and needs to be
    installed before using the function.

    :param result_data: An input representation of (Batch) Statistical API result returned from
        `AwsBatchStatisticalResults.get_data()`. Each JsonDict in the list is a Statistical API response of an input
        geometry.
    :param exclude_stats: The statistic names defined in this parameter will be excluded from the output DataFrame.

    :return: Statistical dataframe.
    """
    try:
        import pandas as pd  # pylint: disable=import-outside-toplevel
    except ImportError as exception:
        raise ImportError(_PANDAS_IMPORT_MESSAGE) from exception

    exclude_stats = exclude_stats or []

    nresults = len(result_data)
    dfs = [None] * nresults
    for idx in range(nresults):
        result = result_data[idx]

        # valid batch stat response
        if _is_batch_stat(result) and _is_valid_batch_response(result):
            identifier, response_data = result["identifier"], result["response"]["data"]

        # valid normal stat response
        elif not _is_batch_stat(result) and "data" in result:
            identifier, response_data = str(idx), result["data"]
        else:
            continue
        result_entries = _extract_response_data(response_data, exclude_stats)
        result_df = pd.DataFrame(result_entries)
        result_df["identifier"] = identifier
        dfs[idx] = result_df
    return pd.concat(dfs)


def _get_failed_intervals(response_data: list[JsonDict]) -> list[tuple[str, str]]:
    """Collect failed intervals of a single geometry from the (Batch) Statistical result

    :param response_data: An input representation of the (Batch) Statistical API response of a geometry.
    :return: The failed intervals for a geometry.
    """
    return [
        (interval["interval"]["from"], interval["interval"]["to"]) for interval in response_data if "error" in interval
    ]


def _get_failed_batch_response(result_data: JsonDict) -> str | list[tuple[str, str]]:
    """Collect failed responses

    :param result_data: An input representation of the (Batch) Statistical API result of a geometry.
    :return: Failed responses and responses with failed intervals
    """
    if "error" in result_data or result_data["response"]["status"] == "FAILED":
        return _FULL_TIME_RANGE
    return _get_failed_intervals(result_data["response"]["data"])


def get_failed_statistical_requests(result_data: list[JsonDict]) -> list[JsonDict]:
    """Collect failed requests of (Batch) Statistical Results

    :param result_data: An input representation of (Batch) Statistical API result.
    :return: Failed requests of (Batch) Statistical Results.
    """
    failed_responses: Iterable[tuple[Any, str | list[tuple[str, str]]]]
    if _is_batch_stat(result_data[0]):
        failed_responses = ((result["identifier"], _get_failed_batch_response(result)) for result in result_data)
    else:
        failed_responses = enumerate(
            _FULL_TIME_RANGE if "error" in result else _get_failed_intervals(result["data"]) for result in result_data
        )
    return [
        {"identifier": idx, "failed_intervals": intervals} for idx, intervals in failed_responses if intervals != []
    ]



"""
Module implementing utilities for collecting data, produced with Sentinel Hub Statistical Batch API, from an S3 bucket.
"""

from typing import List, Optional, Sequence, Union

from ..api.batch.statistical import BatchStatisticalRequest, BatchStatisticalRequestType, SentinelHubBatchStatistical
from ..base import DataRequest
from ..config import SHConfig
from ..constants import MimeType
from ..download.models import DownloadRequest
from .client import AwsDownloadClient


class AwsBatchStatisticalResults(DataRequest):
    """A utility class for downloading results of Batch Statistical API from an S3 bucket."""

    def __init__(
        self,
        batch_request: BatchStatisticalRequestType,
        *,
        feature_ids: Optional[Sequence[Union[str, int]]] = None,
        data_folder: Optional[str] = None,
        config: Optional[SHConfig] = None,
    ):
        """
        :param batch_request: Info about a batch request - either an instance of `BatchStatisticalRequest` or a
            batch ID or a raw payload of the batch response.
        :param feature_ids: A list of feature IDs of saved results on the bucket. If provided it will download only
            these results. If not provided it will collect the names of all JSON files from results folder on the
            bucket and download all of them. Note that it is recommended that you provide this parameter otherwise this
            class will have to make additional requests to the S3 bucket in order to list all features from the folder.
        :param data_folder: Directory to which the files should be saved.
        :param config: A config object that contains AWS credentials to access the S3 bucket with results.
        """
        self.batch_request = self._parse_batch_request(batch_request, config)
        self.feature_ids = feature_ids

        super().__init__(AwsDownloadClient, data_folder=data_folder, config=config)

    @staticmethod
    def _parse_batch_request(
        batch_request: BatchStatisticalRequestType, config: Optional[SHConfig]
    ) -> BatchStatisticalRequest:
        """In case a batch request is not defined with an instance of `BatchStatisticalRequest` it will make sure that
        such an instance is created."""
        if isinstance(batch_request, BatchStatisticalRequest):
            return batch_request

        if isinstance(batch_request, dict):
            return BatchStatisticalRequest.from_dict(batch_request)

        batch_client = SentinelHubBatchStatistical(config=config)
        return batch_client.get_request(batch_request)

    def create_request(self) -> None:
        """Creates a list of download requests."""
        base_s3_path = self.batch_request.request["output"]["s3"]["url"].rstrip("/")
        s3_path = f"{base_s3_path}/{self.batch_request.request_id}/"

        filenames = self._get_filenames(s3_path)

        self.download_list = [
            DownloadRequest(
                url=f"{s3_path}{filename}", data_folder=self.data_folder, data_type=MimeType.JSON, filename=filename
            )
            for filename in filenames
        ]

    def _get_filenames(self, s3_path: str) -> List[str]:
        """Creates a list of JSON filenames from given feature ids or from given S3 path if feature ids are not
        provided. In case if it has to collect them from S3 path it makes sure not to collect any data from any
        subfolder in the path."""
        if self.feature_ids is not None:
            return [f"{feature_id}.json" for feature_id in self.feature_ids]

        filenames: List[str] = []

        s3_client = AwsDownloadClient.get_s3_client(self.config)
        _, _, bucket_name, url_key = s3_path.split("/", 3)

        paginator = s3_client.get_paginator("list_objects")
        for page in paginator.paginate(Bucket=bucket_name, Prefix=url_key):
            for item in page["Contents"]:
                key_path = item["Key"]
                key_name = key_path.rsplit("/", 1)[1]

                if key_name.endswith(".json") and key_path == f"{url_key}{key_name}":
                    filenames.append(key_name)

        return filenames



"""
Module implementing some utility functions not suitable for other utility modules
"""

# ruff: noqa: FA100
# do not use `from __future__ import annotations`, it clashes with `dataclass_json`
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, Iterable, Optional, Protocol, Union
from urllib.parse import urlencode

from dataclasses_json import CatchAll, LetterCase, Undefined, dataclass_json
from dataclasses_json import config as dataclass_config

from ..base import FeatureIterator
from ..config import SHConfig
from ..data_collections import DataCollection
from ..download.sentinelhub_client import SentinelHubDownloadClient
from ..exceptions import MissingDataInRequestException
from ..types import JsonDict
from .utils import datetime_config, remove_undefined


class SentinelHubService(metaclass=ABCMeta):
    """A base class for classes interacting with different Sentinel Hub APIs"""

    _DEFAULT_RETRY_TIME = 30

    def __init__(self, config: Optional[SHConfig] = None):
        """
        :param config: A configuration object with required parameters `sh_client_id`, `sh_client_secret`, and
            `sh_auth_base_url` which is used for authentication and `sh_base_url` which defines the service
            deployment that will be used.
        """
        self.config = config or SHConfig()

        base_url = self.config.sh_base_url.rstrip("/")
        self.service_url = self._get_service_url(base_url)

        self.client = SentinelHubDownloadClient(config=self.config, default_retry_time=self._DEFAULT_RETRY_TIME)

    @staticmethod
    @abstractmethod
    def _get_service_url(base_url: str) -> str:
        """Provides the URL to a specific service"""


class SentinelHubFeatureIterator(FeatureIterator[JsonDict]):
    """Feature iterator for the most common implementation of feature pagination at Sentinel Hub services"""

    def __init__(self, *args: Any, exception_message: Optional[str] = None, **kwargs: Any):
        """
        :param args: Arguments passed to FeatureIterator
        :param exception_message: A message to be raised if no features are found
        :param kwargs: Keyword arguments passed to FeatureIterator
        """
        self.exception_message = exception_message or "No data found"
        self.next: Optional[JsonDict] = None

        super().__init__(*args, **kwargs)

    def _fetch_features(self) -> Iterable[JsonDict]:
        """Collect more results from the service"""
        params = remove_undefined({**self.params, "viewtoken": self.next})
        url = f"{self.url}?{urlencode(params)}"

        json_response = self.client.get_json_dict(url, use_session=True)

        new_features = json_response.get("data")
        if new_features is None:
            raise MissingDataInRequestException(self.exception_message)

        self.next = json_response.get("links", {}).get("nextToken")
        self.finished = self.next is None or not new_features

        return new_features


class _AdditionalData(Protocol):
    """Describes minimum requirements for additional data passed to BaseCollection"""

    bands: Optional[Dict[str, Any]]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class BaseCollection:
    """Dataclass to hold data about a collection"""

    name: str
    s3_bucket: str
    additional_data: Optional[_AdditionalData]
    collection_id: Optional[str] = field(metadata=dataclass_config(field_name="id"), default=None)
    user_id: Optional[str] = None
    created: Optional[datetime] = field(metadata=datetime_config, default=None)
    no_data: Optional[Union[int, float]] = None
    other_data: CatchAll = field(default_factory=dict)

    def to_data_collection(self) -> DataCollection:
        """Returns a DataCollection enum for this collection"""
        if self.collection_id is None:
            raise ValueError("This collection is missing a collection id")

        if self.additional_data and self.additional_data.bands:
            band_names = tuple(self.additional_data.bands)
        else:
            band_names = None

        return DataCollection.define_byoc(collection_id=self.collection_id, bands=band_names)



"""
The part of the package that implements interface with Sentinel Hub services.
"""

from .batch import (
    BatchProcessClient,
    BatchProcessRequest,
    BatchRequestStatus,
    BatchStatisticalRequest,
    BatchUserAction,
    SentinelHubBatchStatistical,
    monitor_batch_process_analysis,
    monitor_batch_process_job,
    monitor_batch_statistical_analysis,
    monitor_batch_statistical_job,
)
from .byoc import ByocCollection, ByocCollectionAdditionalData, ByocCollectionBand, ByocTile, SentinelHubBYOC
from .catalog import SentinelHubCatalog
from .ogc import WcsRequest, WmsRequest
from .process import AsyncProcessRequest, SentinelHubRequest, get_async_running_status
from .statistical import SentinelHubStatistical
from .wfs import WebFeatureService



"""
Implementation of `Sentinel Hub Process API interface <https://docs.sentinel-hub.com/api/latest/api/process/>`__.
"""

from __future__ import annotations

from typing import Any, Iterable

import requests

from sentinelhub.exceptions import DownloadFailedException

from ..config import SHConfig
from ..constants import MimeType
from ..download import SentinelHubDownloadClient
from ..geometry import BBox, Geometry
from ..types import JsonDict
from .base_request import InputDataDict, SentinelHubBaseApiRequest
from .utils import AccessSpecification, _update_other_args, remove_undefined, s3_specification


class SentinelHubRequest(SentinelHubBaseApiRequest):
    """Sentinel Hub Process API interface

    For more information check
    `Process API documentation <https://docs.sentinel-hub.com/api/latest/api/process/>`__.
    """

    _SERVICE_ENDPOINT = "process"

    def __init__(
        self,
        evalscript: str,
        input_data: list[JsonDict | InputDataDict],
        responses: list[JsonDict],
        bbox: BBox | None = None,
        geometry: Geometry | None = None,
        size: tuple[int, int] | None = None,
        resolution: tuple[float, float] | None = None,
        **kwargs: Any,
    ):
        """
        For details of certain parameters check the
        `Process API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/process>`_.

        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :param input_data: A list of input dictionary objects as described in the API reference. It can be generated
            with the helper method `SentinelHubRequest.input_data`
        :param responses: A list of `output.responses` objects as described in the API reference. It can be generated
            with the helper function `SentinelHubRequest.output_response`
        :param bbox: Bounding box describing the area of interest.
        :param geometry: Geometry describing the area of interest.
        :param size: Size of the image.
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        if not isinstance(evalscript, str):
            raise ValueError("'evalscript' should be a string")

        parsed_mime_type = MimeType.from_string(responses[0]["format"]["type"])
        self._mime_type = MimeType.TAR if len(responses) > 1 else parsed_mime_type

        self.payload = self.body(
            request_bounds=self.bounds(bbox=bbox, geometry=geometry),
            request_data=input_data,
            request_output=self.output(size=size, resolution=resolution, responses=responses),
            evalscript=evalscript,
        )

        super().__init__(SentinelHubDownloadClient, **kwargs)

    @property
    def mime_type(self) -> MimeType:
        return self._mime_type

    @staticmethod
    def body(
        request_bounds: JsonDict,
        request_data: list[JsonDict],
        evalscript: str,
        request_output: JsonDict | None = None,
        other_args: JsonDict | None = None,
    ) -> JsonDict:
        """Generate the Process API request body

        :param request_bounds: A dictionary as generated by `SentinelHubRequest.bounds` helper method.
        :param request_data: A list of dictionaries as generated by `SentinelHubRequest.input_data` helper method.
        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :param request_output: A dictionary as generated by `SentinelHubRequest.output` helper method.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        request_body = {"input": {"bounds": request_bounds, "data": request_data}, "evalscript": evalscript}

        if request_output is not None:
            request_body["output"] = request_output

        if other_args:
            _update_other_args(request_body, other_args)

        return request_body

    @staticmethod
    def output_response(
        identifier: str, response_format: str | MimeType, other_args: JsonDict | None = None
    ) -> JsonDict:
        """Generate an element of `output.responses` as described in the Process API reference.

        :param identifier: Identifier of the output response.
        :param response_format: A mime type of one of 'png', 'json', 'jpeg', 'tiff'.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        output_response = {"identifier": identifier, "format": {"type": MimeType(response_format).get_string()}}

        if other_args:
            _update_other_args(output_response, other_args)

        return output_response

    @staticmethod
    def output(
        responses: list[JsonDict],
        size: tuple[int, int] | None = None,
        resolution: tuple[float, float] | None = None,
        other_args: dict | None = None,
    ) -> JsonDict:
        """Generate an `output` part of the request as described in the Process API reference

        :param responses: A list of objects in `output.responses` as generated by `SentinelHubRequest.output_response`.
        :param size: Size of the image.
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        if size and resolution:
            raise ValueError("Either size or resolution argument should be given, not both.")

        request_output: JsonDict = {"responses": responses}

        if size:
            request_output["width"], request_output["height"] = size
        if resolution:
            request_output["resx"], request_output["resy"] = resolution

        if other_args:
            _update_other_args(request_output, other_args)

        return request_output


class AsyncProcessRequest(SentinelHubBaseApiRequest):
    """Sentinel Hub Async Process API interface

    For more information check
    `Async Process API documentation <https://docs.sentinel-hub.com/api/latest/api/async-process/>`__.
    """

    _SERVICE_ENDPOINT = "async/process"

    def __init__(
        self,
        *,
        evalscript: str | None = None,
        evalscript_reference: AccessSpecification | None = None,
        input_data: list[JsonDict | InputDataDict],
        responses: list[JsonDict],
        delivery: AccessSpecification,
        bbox: BBox | None = None,
        geometry: Geometry | None = None,
        size: tuple[int, int] | None = None,
        resolution: tuple[float, float] | None = None,
        **kwargs: Any,
    ):
        """
        For details of certain parameters check the
        `Async Process API reference <https://docs.sentinel-hub.com/api/latest/reference/#tag/async_process>`_.

        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :param evalscript_reference:
            `Evalscript specification <https://docs.sentinel-hub.com/api/latest/reference/#tag/async_process>`__.
        :param input_data: A list of input dictionary objects as described in the API reference. It can be generated
            with the helper method `SentinelHubRequest.input_data`
        :param responses: A list of `output.responses` objects as described in the API reference. It can be generated
            with the helper function `SentinelHubRequest.output_response`
        :param delivery: S3 specification for dumping the request data.
        :param bbox: Bounding box describing the area of interest.
        :param geometry: Geometry describing the area of interest.
        :param size: Size of the image.
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """

        self._mime_type = MimeType.JSON

        self.payload = self.body(
            request_bounds=self.bounds(bbox=bbox, geometry=geometry),
            request_data=input_data,
            request_output=self.output(size=size, resolution=resolution, responses=responses, delivery=delivery),
            evalscript=evalscript,
            evalscript_reference=evalscript_reference,
        )

        super().__init__(SentinelHubDownloadClient, **kwargs)

    s3_specification = staticmethod(s3_specification)

    @property
    def mime_type(self) -> MimeType:
        return self._mime_type

    @staticmethod
    def body(
        request_bounds: JsonDict,
        request_data: list[JsonDict],
        evalscript: str | None,
        evalscript_reference: AccessSpecification | None,
        request_output: JsonDict | None = None,
        other_args: JsonDict | None = None,
    ) -> JsonDict:
        """Generate the Async Process API request body

        :param request_bounds: A dictionary as generated by `SentinelHubRequest.bounds` helper method.
        :param request_data: A list of dictionaries as generated by `SentinelHubRequest.input_data` helper method.
        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :param request_output: A dictionary as generated by `SentinelHubRequest.output` helper method.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        request_body = {
            "input": {"bounds": request_bounds, "data": request_data},
            "evalscript": evalscript,
            "evalscriptReference": evalscript_reference,
        }

        if request_output is not None:
            request_body["output"] = request_output

        if other_args:
            _update_other_args(request_body, other_args)

        return remove_undefined(request_body)

    @staticmethod
    def output_response(
        identifier: str, response_format: str | MimeType, other_args: JsonDict | None = None
    ) -> JsonDict:
        """Generate an element of `output.responses` as described in the Async Process API reference.

        :param identifier: Identifier of the output response.
        :param response_format: A mime type of one of 'png', 'json', 'jpeg', 'tiff'.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        output_response = {"identifier": identifier, "format": {"type": MimeType(response_format).get_string()}}

        if other_args:
            _update_other_args(output_response, other_args)

        return output_response

    @staticmethod
    def output(
        responses: list[JsonDict],
        delivery: AccessSpecification,
        size: tuple[int, int] | None = None,
        resolution: tuple[float, float] | None = None,
        other_args: dict | None = None,
    ) -> JsonDict:
        """Generate an `output` part of the request as described in the Async Process API reference

        :param responses: A list of objects in `output.responses` as generated by `AsyncProcessRequest.output_response`.
        :param size: Size of the image.
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        if size and resolution:
            raise ValueError("Either size or resolution argument should be given, not both.")

        request_output: JsonDict = {"responses": responses, "delivery": delivery}

        if size:
            request_output["width"], request_output["height"] = size
        if resolution:
            request_output["resx"], request_output["resy"] = resolution

        if other_args:
            _update_other_args(request_output, other_args)

        return request_output


def get_async_running_status(ids: Iterable[str], config: SHConfig | None = None) -> dict[str, bool]:
    """Returns a mapping that describes which requests are running.

    :param ids: A collection of async request IDs.
    :param config: A custom instance of config class to override parameters from the saved configuration.
    :return: A mapping that specifies whether a process is running for each of the IDs.
    """
    config = config or SHConfig()
    client = SentinelHubDownloadClient(config=config)
    result = {}
    for request_id in ids:
        try:
            client.get_json_dict(f"{config.sh_base_url}/api/v1/async/process/{request_id}", use_session=True)
            # A successful request means it's running
            result[request_id] = True
        except DownloadFailedException as exception:  # noqa: PERF203
            # A 404 means it's not running
            if exception.request_exception is not None and exception.request_exception.response is not None:
                if exception.request_exception.response.status_code == requests.status_codes.codes.NOT_FOUND:
                    result[request_id] = False
            else:
                raise exception from exception

    return result



"""
Module containing shared code of Batch Process API and Batch Statistical API
"""

# ruff: noqa: FA100
# do not use `from __future__ import annotations`, it clashes with `dataclass_json` (even through inheritance)
from abc import ABCMeta
from enum import Enum
from typing import Generic, Iterable, Optional, Sequence, Type, TypeVar, Union

from ...constants import RequestType
from ...types import Json, JsonDict
from ..base import SentinelHubService

BatchRequestType = TypeVar("BatchRequestType", bound="BaseBatchRequest")  # pylint: disable=invalid-name
RequestSpec = Union[str, dict, BatchRequestType]
Self = TypeVar("Self")


class BatchRequestStatus(Enum):
    """An enum class with all possible batch request statuses"""

    CREATED = "CREATED"
    ANALYSING = "ANALYSING"
    ANALYSIS_DONE = "ANALYSIS_DONE"
    PROCESSING = "PROCESSING"
    DONE = "DONE"
    FAILED = "FAILED"
    STOPPED = "STOPPED"


class BatchUserAction(Enum):
    """An enum class with all possible batch user actions"""

    START = "START"
    ANALYSE = "ANALYSE"
    STOP = "STOP"
    NONE = "NONE"


class StoppedStatusReason(Enum):
    """Description of why job status is STOPPED"""

    OUT_OF_PU = "OUT_OF_PU"
    USER_ACTION = "USER_ACTION"
    UNHEALTHY = "UNHEALTHY"


class BaseBatchClient(SentinelHubService, Generic[BatchRequestType], metaclass=ABCMeta):
    """Class containing common methods and helper functions for Batch Client classes"""

    def _call_job(self, batch_request: RequestSpec, endpoint_name: str) -> Json:
        """Makes a POST request to the service that triggers a processing job"""
        request_id = self._parse_request_id(batch_request)
        job_url = f"{self._get_processing_url(request_id)}/{endpoint_name}"
        return self.client.get_json(url=job_url, request_type=RequestType.POST, use_session=True)

    def _get_processing_url(self, request_id: Optional[str] = None) -> str:
        """Creates a URL for the batch statistical endpoint"""
        url = self.service_url
        if request_id is None:
            return url
        return f"{url}/{request_id}"

    @staticmethod
    def _parse_request_id(data: RequestSpec) -> str:
        """Parses batch request id from multiple possible inputs"""
        if isinstance(data, BaseBatchRequest):
            return data.request_id
        if isinstance(data, dict):
            return data["id"]
        if isinstance(data, str):
            return data
        raise ValueError(f"Expected a BatchRequest, dictionary or a string, got {data}.")


class BaseBatchRequest:
    """Class containing helper functions for Batch Request classes"""

    _REPR_PARAM_NAMES: Sequence[str]

    request_id: str
    error: Optional[str]
    status: BatchRequestStatus

    def to_dict(self) -> JsonDict:
        """Transforms itself into a dictionary form."""
        raise NotImplementedError("Method should be implemented or provided via `dataclass_json` decorator.")

    @classmethod
    def from_dict(cls: Type[Self], json_dict: JsonDict) -> Self:
        """Transforms itself into a dictionary form."""
        raise NotImplementedError("Method should be implemented or provided via `dataclass_json` decorator.")

    def __repr__(self) -> str:
        """A representation that shows the basic parameters of a batch job"""
        repr_params = {name: getattr(self, name) for name in self._REPR_PARAM_NAMES if getattr(self, name) is not None}
        repr_params_str = "\n  ".join(f"{name}={value}" for name, value in repr_params.items())
        return f"{self.__class__.__name__}(\n  {repr_params_str}\n  ...\n)"

    def raise_for_status(
        self,
        status: Union[str, BatchRequestStatus, Iterable[Union[str, BatchRequestStatus]]] = BatchRequestStatus.FAILED,
    ) -> None:
        """Raises an error in case batch request has a given status

        :param status: One or more status codes on which to raise an error. The default is `'FAILED'`.
        :raises: RuntimeError
        """
        if isinstance(status, (str, BatchRequestStatus)):
            status = [status]
        status_list = [BatchRequestStatus(_status) for _status in status]

        if self.status in status_list:
            formatted_error_message = f' and error message: "{self.error}"' if self.error else ""
            raise RuntimeError(
                f"Raised for batch request {self.request_id} with status {self.status.value}{formatted_error_message}"
            )



"""Module with custom types and utilities used in sentinelhub-py."""

import datetime as dt
from typing import Any, Dict, Tuple, Union

from typing_extensions import TypeAlias

RawTimeType: TypeAlias = Union[None, str, dt.date]
RawTimeIntervalType: TypeAlias = Tuple[RawTimeType, RawTimeType]
JsonDict: TypeAlias = Dict[str, Any]
Json: TypeAlias = Union[JsonDict, list, str, float, int, None]



"""
Module for managing configuration data from `config.toml`
"""

from __future__ import annotations

import copy
import json
import os
import warnings
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any

import tomli
import tomli_w

from .exceptions import SHDeprecationWarning

DEFAULT_PROFILE = "default-profile"
SH_PROFILE_ENV_VAR = "SH_PROFILE"
SH_CLIENT_ID_ENV_VAR = "SH_CLIENT_ID"
SH_CLIENT_SECRET_ENV_VAR = "SH_CLIENT_SECRET"


@dataclass(repr=False)
class _SHConfig:  # pylint: disable=too-many-instance-attributes
    instance_id: str = ""
    sh_client_id: str = ""
    sh_client_secret: str = ""
    sh_base_url: str = "https://services.sentinel-hub.com"
    sh_auth_base_url: str | None = None
    sh_token_url: str = "https://services.sentinel-hub.com/auth/realms/main/protocol/openid-connect/token"
    geopedia_wms_url: str = "https://service.geopedia.world"
    geopedia_rest_url: str = "https://www.geopedia.world/rest"
    aws_access_key_id: str = ""
    aws_secret_access_key: str = ""
    aws_session_token: str = ""
    aws_metadata_url: str = "https://roda.sentinel-hub.com"
    aws_s3_l1c_bucket: str = "sentinel-s2-l1c"
    aws_s3_l2a_bucket: str = "sentinel-s2-l2a"
    opensearch_url: str = "http://opensearch.sentinel-hub.com/resto/api/collections/Sentinel2"
    max_wfs_records_per_query: int = 100
    max_opensearch_records_per_query: int = 500  # pylint: disable=invalid-name
    max_download_attempts: int = 4
    download_sleep_time: float = 5.0
    download_timeout_seconds: float = 120.0
    number_of_download_processes: int = 1
    max_retries: int | None = None

    def __post_init__(self) -> None:
        if self.sh_auth_base_url is not None:
            self.sh_token_url = self.sh_auth_base_url + "/oauth/token"
            warnings.warn(
                "The parameter `sh_auth_base_url` of `SHConfig` has been replaced with `sh_token_url`. Please"
                " update your configuration, for now the parameters were automatically adjusted to `sh_token_url ="
                " sh_auth_base_url + '/oauth/token'`.",
                category=SHDeprecationWarning,
            )

        if self.max_wfs_records_per_query > 100:
            raise ValueError("Value of config parameter `max_wfs_records_per_query` must be at most 100")
        if self.max_opensearch_records_per_query > 500:
            raise ValueError("Value of config parameter `max_opensearch_records_per_query` must be at most 500")


class SHConfig(_SHConfig):
    """A sentinelhub-py package configuration class.

    The class reads the configurable settings from ``config.toml`` file on initialization:

        - `instance_id`: An instance ID for Sentinel Hub service used for OGC requests.
        - `sh_client_id`: User's OAuth client ID for Sentinel Hub service. Can be set via SH_CLIENT_ID environment
          variable. The environment variable has precedence.
        - `sh_client_secret`: User's OAuth client secret for Sentinel Hub service. Can be set via SH_CLIENT_SECRET
          environment variable. The environment variable has precedence.
        - `sh_base_url`: There exist multiple deployed instances of Sentinel Hub service, this parameter defines the
          location of a specific service instance.
        - `sh_token_url`: Url for Sentinel Hub Authentication service. Authentication is typically sent to the main
          service deployment even if `sh_base_url` points to another deployment.
        - `geopedia_wms_url`: Base url for Geopedia WMS services.
        - `geopedia_rest_url`: Base url for Geopedia REST services.
        - `aws_access_key_id`: Access key for AWS Requester Pays buckets.
        - `aws_secret_access_key`: Secret access key for AWS Requester Pays buckets.
        - `aws_session_token`: A session token for your AWS account. It is only needed when you are using temporary
          credentials.
        - `aws_metadata_url`: Base url for publicly available metadata files
        - `aws_s3_l1c_bucket`: Name of Sentinel-2 L1C bucket at AWS s3 service.
        - `aws_s3_l2a_bucket`: Name of Sentinel-2 L2A bucket at AWS s3 service.
        - `opensearch_url`: Base url for Sentinelhub Opensearch service.
        - `max_wfs_records_per_query`: Maximum number of records returned for each WFS query.
        - `max_opensearch_records_per_query`: Maximum number of records returned for each Opensearch query.
        - `max_download_attempts`: Maximum number of download attempts from a single URL until an error will be raised.
        - `download_sleep_time`: Number of seconds to sleep between the first failed attempt and the next. Every next
          attempt this number exponentially increases with factor `3`.
        - `download_timeout_seconds`: Maximum number of seconds before download attempt is canceled.
        - `number_of_download_processes`: Number of download processes, used to calculate rate-limit sleep time.
        - `max_retries`: Maximum number of retries until an exception is raised.

    The location of `config.toml` for manual modification can be found with `SHConfig.get_config_location()`.
    """

    CREDENTIALS = (
        "instance_id",
        "sh_client_id",
        "sh_client_secret",
        "aws_access_key_id",
        "aws_secret_access_key",
        "aws_session_token",
    )

    def __init__(self, profile: str | None = None, *, use_defaults: bool = False, **kwargs: Any):
        """
        :param profile: Specifies which profile to load from the configuration file. Has precedence over the environment
            variable `SH_USER_PROFILE`.
        :param use_defaults: Does not load the configuration file, returns config object with defaults only.
        :param kwargs: Any fields of `SHConfig` to be updated. Overrides settings from `config.toml` and environment.
        """
        profile = self._get_profile(profile)

        if not use_defaults:
            env_kwargs = {
                "sh_client_id": os.environ.get(SH_CLIENT_ID_ENV_VAR),
                "sh_client_secret": os.environ.get(SH_CLIENT_SECRET_ENV_VAR),
            }
            env_kwargs = {k: v for k, v in env_kwargs.items() if v is not None}

            # load from config.toml
            loaded_kwargs = SHConfig.load(profile=profile).to_dict(mask_credentials=False)

            kwargs = {**loaded_kwargs, **env_kwargs, **kwargs}  # precedence: init params > env > loaded

        super().__init__(**kwargs)

    def __str__(self) -> str:
        """Content of `SHConfig` in json schema. Credentials are masked for safety."""
        return json.dumps(self.to_dict(mask_credentials=True), indent=2)

    def __repr__(self) -> str:
        """Representation of `SHConfig`. Credentials are masked for safety."""
        config_dict = self.to_dict(mask_credentials=True)
        content = ",\n  ".join(f"{key}={value!r}" for key, value in config_dict.items())
        return f"{self.__class__.__name__}(\n  {content},\n)"

    @staticmethod
    def _get_profile(profile: str | None) -> str:
        return profile if profile is not None else os.environ.get(SH_PROFILE_ENV_VAR, default=DEFAULT_PROFILE)

    @classmethod
    def load(cls, profile: str | None = None) -> SHConfig:
        """Loads configuration parameters from the config file at `SHConfig.get_config_location()`.

        :param profile: Which profile to load from the configuration file.
        """
        profile = cls._get_profile(profile)
        filename = cls.get_config_location()
        if not os.path.exists(filename):
            cls(use_defaults=True).save()  # store default configuration to standard location

        with open(filename, "rb") as cfg_file:
            configurations_dict = tomli.load(cfg_file)

        if profile not in configurations_dict:
            raise KeyError(f"Profile `{profile}` not found in configuration file.")

        return cls(use_defaults=True, **configurations_dict[profile])

    def save(self, profile: str | None = None) -> None:
        """Saves configuration parameters to the config file at `SHConfig.get_config_location()`.

        :param profile: Under which profile to save the configuration.
        """
        profile = self._get_profile(profile)
        file_path = Path(self.get_config_location())
        file_path.parent.mkdir(parents=True, exist_ok=True)

        if file_path.exists():
            with open(file_path, "rb") as cfg_file:
                current_configuration = tomli.load(cfg_file)
        else:
            current_configuration = {}

        current_configuration[profile] = self._get_dict_of_diffs_from_defaults()
        with open(file_path, "wb") as cfg_file:
            tomli_w.dump(current_configuration, cfg_file)

    def _get_dict_of_diffs_from_defaults(self) -> dict[str, str | float]:
        """Returns a dictionary containing key: value pairs for parameters that have values different from defaults."""
        current_profile_config = self.to_dict(mask_credentials=False)
        default_values = SHConfig(use_defaults=True).to_dict(mask_credentials=False)
        return {key: value for key, value in current_profile_config.items() if default_values[key] != value}

    def copy(self) -> SHConfig:
        """Makes a copy of an instance of `SHConfig`"""
        return copy.copy(self)

    def to_dict(self, mask_credentials: bool = True) -> dict[str, str | float]:
        """Get a dictionary representation of the `SHConfig` class.

        :param mask_credentials: Wether to mask fields containing credentials.
        :return: A dictionary with configuration parameters
        """
        config_params = asdict(self)

        if mask_credentials:
            for param in self.CREDENTIALS:
                config_params[param] = self._mask_credentials(config_params[param])

        return config_params

    def _mask_credentials(self, value: str) -> str:
        """In case a parameter that holds credentials is given it will mask its value"""
        hide_size = min(max(len(value) - 4, 10), len(value))
        return "*" * hide_size + value[hide_size:]

    @classmethod
    def get_config_location(cls) -> str:
        """Returns the default location of the user configuration file on disk."""
        config_folder = os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))
        return os.path.join(config_folder, "sentinelhub", "config.toml")



"""
Module implementing utilities for working with batch jobs.
"""

from __future__ import annotations

import logging
import time
from typing import Union

from tqdm.auto import tqdm

from ...config import SHConfig
from ...types import JsonDict
from .base import BatchRequestStatus
from .process import BatchProcessClient, BatchProcessRequest
from .statistical import BatchStatisticalRequest, SentinelHubBatchStatistical

LOGGER = logging.getLogger(__name__)

BatchStatisticalRequestSpec = Union[str, dict, BatchStatisticalRequest]


_MIN_SLEEP_TIME = 60
_DEFAULT_SLEEP_TIME = 120
_MIN_STAT_SLEEP_TIME = 15
_DEFAULT_STAT_SLEEP_TIME = 30
_MIN_ANALYSIS_SLEEP_TIME = 5
_DEFAULT_ANALYSIS_SLEEP_TIME = 10


def monitor_batch_process_job(
    request: BatchProcessRequest,
    client: BatchProcessClient,
    sleep_time: int = _DEFAULT_SLEEP_TIME,
    analysis_sleep_time: int = _DEFAULT_ANALYSIS_SLEEP_TIME,
) -> BatchProcessRequest:
    """A utility function that keeps checking the progress of the batch processing job. Returns an updated version of
    the request

    Notes:

      - Before calling this function make sure to start a batch job by calling `BatchProcessingClient.start_job` method.
        In case a batch job is still being analysed this function will wait until the analysis ends.
      - This function will be continuously collecting information from Sentinel Hub service. To avoid making too many
        requests please make sure to adjust `sleep_time` parameter.

    :param request: The request to monitor.
    :param client: A batch processing client with appropriate configuration that is used to monitor the batch job.
    :param sleep_time: Number of seconds to sleep between consecutive progress bar updates.
    :param analysis_sleep_time: Number of seconds between consecutive status updates during analysis phase.
    """
    if sleep_time < _MIN_SLEEP_TIME:
        raise ValueError(f"To avoid making too many service requests please set sleep_time>={_MIN_SLEEP_TIME}")

    batch_request: BatchProcessRequest = monitor_batch_process_analysis(request, client, sleep_time=analysis_sleep_time)
    if batch_request.status is BatchRequestStatus.PROCESSING:
        LOGGER.info("Batch job is running")

    completion = batch_request.completion_percentage
    progress_bar = tqdm(total=100, initial=completion, desc="Completion percentage")

    monitoring_status = [BatchRequestStatus.ANALYSIS_DONE, BatchRequestStatus.PROCESSING]
    with progress_bar:
        while completion < 100 and batch_request.status in monitoring_status:
            time.sleep(sleep_time)
            batch_request = client.get_request(batch_request)
            progress_bar.update(batch_request.completion_percentage - completion)
            completion = batch_request.completion_percentage

    while batch_request.status in monitoring_status:
        LOGGER.info("Waiting on batch job status update, currently %s", batch_request.status)
        time.sleep(sleep_time)
        batch_request = client.get_request(batch_request)

    LOGGER.info("Batch job finished with status %s", batch_request.status.value)
    return batch_request


def monitor_batch_statistical_job(
    batch_request: BatchStatisticalRequestSpec,
    config: SHConfig | None = None,
    sleep_time: int = _DEFAULT_STAT_SLEEP_TIME,
    analysis_sleep_time: int = _DEFAULT_ANALYSIS_SLEEP_TIME,
) -> JsonDict:
    """A utility function that keeps checking the completion percentage of a Batch Statistical request until complete.

    Notes:

      - Before calling this function make sure to start a batch job via `SentinelHubBatchStatistical.start_job` method.
        In case a batch job is still being analysed this function will wait until the analysis ends.
      - Some information about the progress of this function is reported to logging level INFO.

    :param batch_request: An object with information about a batch request. Alternatively, it could only be a batch
        request id or a payload.
    :param config: A configuration object with required parameters `sh_client_id`, `sh_client_secret`, and
        `sh_auth_base_url` which is used for authentication and `sh_base_url` which defines the service deployment
        where Batch API will be called.
    :param sleep_time: Number of seconds to sleep between consecutive progress bar updates.
    :param analysis_sleep_time: Number of seconds between consecutive status updates during analysis phase.
    :return: Final status of the batch request.
    """
    if sleep_time < _MIN_STAT_SLEEP_TIME:
        raise ValueError(f"To avoid making too many service requests please set sleep_time>={_MIN_STAT_SLEEP_TIME}")

    batch_request = monitor_batch_statistical_analysis(batch_request, config, sleep_time=analysis_sleep_time)
    if batch_request.status is BatchRequestStatus.PROCESSING:
        LOGGER.info("Batch job is running")

    batch_client = SentinelHubBatchStatistical(config=config)

    request_status = batch_client.get_status(batch_request)
    progress = request_status["completionPercentage"]
    with tqdm(total=100, initial=progress, desc="Completion percentage") as progress_bar:
        while progress < 100:
            time.sleep(sleep_time)

            request_status = batch_client.get_status(batch_request)
            new_progress = request_status["completionPercentage"]
            progress_bar.update(new_progress - progress)
            progress = new_progress
    return request_status


def monitor_batch_process_analysis(
    request: BatchProcessRequest,
    client: BatchProcessClient,
    sleep_time: int = _DEFAULT_ANALYSIS_SLEEP_TIME,
) -> BatchProcessRequest:
    """A utility function that is waiting until analysis phase of a batch job finishes and regularly checks its status.
    In case analysis phase failed it raises an error at the end.

    :param request: The request to monitor.
    :param client: A batch processing client with appropriate configuration that is used to monitor the batch job.
    :param sleep_time: Number of seconds between consecutive status updates during analysis phase.
    """
    if sleep_time < _MIN_ANALYSIS_SLEEP_TIME:
        raise ValueError(
            f"To avoid making too many service requests please set analysis sleep time >={_MIN_ANALYSIS_SLEEP_TIME}"
        )

    batch_request = client.get_request(request)
    while batch_request.status in [BatchRequestStatus.CREATED, BatchRequestStatus.ANALYSING]:
        LOGGER.info("Batch job has a status %s, sleeping for %d seconds", batch_request.status.value, sleep_time)
        time.sleep(sleep_time)
        batch_request = client.get_request(batch_request)

    batch_request.raise_for_status(status=[BatchRequestStatus.FAILED, BatchRequestStatus.STOPPED])
    return batch_request


def monitor_batch_statistical_analysis(
    batch_request: BatchStatisticalRequestSpec,
    config: SHConfig | None = None,
    sleep_time: int = _DEFAULT_ANALYSIS_SLEEP_TIME,
) -> BatchStatisticalRequest:
    """A utility function that is waiting until analysis phase of a batch job finishes and regularly checks its status.
    In case analysis phase failed it raises an error at the end.

    :param batch_request: An object with information about a batch request. Alternatively, it could only be a batch
        request id or a payload.
    :param config: A configuration object with required parameters `sh_client_id`, `sh_client_secret`, and
        `sh_auth_base_url` which is used for authentication and `sh_base_url` which defines the service deployment
        where Batch API will be called.
    :param sleep_time: Number of seconds between consecutive status updates during analysis phase.
    :return: Batch request info
    """
    if sleep_time < _MIN_ANALYSIS_SLEEP_TIME:
        raise ValueError(
            f"To avoid making too many service requests please set analysis sleep time >={_MIN_ANALYSIS_SLEEP_TIME}"
        )

    batch_client = SentinelHubBatchStatistical(config=config)
    request_status = BatchRequestStatus(batch_client.get_status(batch_request)["status"])
    while request_status in [BatchRequestStatus.CREATED, BatchRequestStatus.ANALYSING]:
        LOGGER.info("Batch job has a status %s, sleeping for %d seconds", request_status.value, sleep_time)
        time.sleep(sleep_time)
        request_status = BatchRequestStatus(batch_client.get_status(batch_request)["status"])

    batch_request = batch_client.get_request(batch_request)
    batch_request.raise_for_status(status=[BatchRequestStatus.FAILED, BatchRequestStatus.STOPPED])
    return batch_request



"""
Module implementing error handlers which can occur during download procedure
"""

from __future__ import annotations

import functools
import logging
import time
from typing import Callable, Protocol, TypeVar

import requests

from ..config import SHConfig
from ..decoding import decode_sentinelhub_err_msg
from ..exceptions import DownloadFailedException
from .models import DownloadRequest


class _HasConfig(Protocol):
    """Interface of objects with a config."""

    config: SHConfig


Self = TypeVar("Self")
SelfWithConfig = TypeVar("SelfWithConfig", bound=_HasConfig)
T = TypeVar("T")


LOGGER = logging.getLogger(__name__)


def fail_user_errors(download_func: Callable[[Self, DownloadRequest], T]) -> Callable[[Self, DownloadRequest], T]:
    """Decorator function for handling user errors"""

    @functools.wraps(download_func)
    def new_download_func(self: Self, request: DownloadRequest) -> T:
        try:
            return download_func(self, request)
        except requests.HTTPError as exception:
            if (
                exception.response.status_code < requests.status_codes.codes.INTERNAL_SERVER_ERROR
                and exception.response.status_code != requests.status_codes.codes.TOO_MANY_REQUESTS
            ):
                raise DownloadFailedException(
                    _create_download_failed_message(exception, request.url), request_exception=exception
                ) from exception
            raise exception from exception

    return new_download_func


def retry_temporary_errors(
    download_func: Callable[[SelfWithConfig, DownloadRequest], T],
) -> Callable[[SelfWithConfig, DownloadRequest], T]:
    """Decorator function for handling server and connection errors"""
    backoff_coefficient = 3

    @functools.wraps(download_func)
    def new_download_func(self: SelfWithConfig, request: DownloadRequest) -> T:
        download_attempts = self.config.max_download_attempts
        sleep_time = self.config.download_sleep_time

        for attempt_idx in range(download_attempts):
            try:
                return download_func(self, request)

            except requests.RequestException as exception:  # noqa: PERF203
                attempts_left = download_attempts - (attempt_idx + 1)
                if not (
                    _is_temporary_problem(exception)
                    or (
                        isinstance(exception, requests.HTTPError)
                        and exception.response.status_code >= requests.status_codes.codes.INTERNAL_SERVER_ERROR
                    )
                ):
                    raise exception from exception

                if attempts_left <= 0:
                    message = _create_download_failed_message(exception, request.url)
                    raise DownloadFailedException(message, request_exception=exception) from exception

                LOGGER.debug(
                    "Download attempt failed: %s\n%d attempts left, will retry in %ds",
                    exception,
                    attempts_left,
                    sleep_time,
                )
                time.sleep(sleep_time)
                sleep_time *= backoff_coefficient

        raise DownloadFailedException(
            "No download attempts available - configuration parameter max_download_attempts should be greater than 0"
        )

    return new_download_func


def fail_missing_file(download_func: Callable[[Self, DownloadRequest], T]) -> Callable[[Self, DownloadRequest], T]:
    """A decorator for raising an error if a file is missing"""

    @functools.wraps(download_func)
    def new_download_func(self: Self, request: DownloadRequest) -> T:
        try:
            return download_func(self, request)
        except requests.HTTPError as exception:
            if exception.response.status_code == requests.status_codes.codes.NOT_FOUND:
                raise DownloadFailedException(
                    f"File in location {request.url} is missing", request_exception=exception
                ) from exception

            raise exception from exception

    return new_download_func


def _is_temporary_problem(exception: Exception) -> bool:
    """Checks if the obtained exception is temporary and if download attempt should be repeated

    :param exception: Exception raised during download
    :return: `True` if exception is temporary and `False` otherwise
    """
    return isinstance(exception, (requests.ConnectionError, requests.Timeout, requests.exceptions.ChunkedEncodingError))


def _create_download_failed_message(exception: Exception, url: str | None) -> str:
    """Creates message describing why download has failed

    :param exception: Exception raised during download
    :param url: A URL from where download was attempted
    :return: Error message
    """
    message = f"Failed to download from:\n{url}\nwith {exception.__class__.__name__}:\n{exception}"

    if _is_temporary_problem(exception):
        if isinstance(exception, requests.ConnectionError):
            message += "\nPlease check your internet connection and try again."
        else:
            message += (
                "\nThere might be a problem in connection or the server failed to process "
                "your request. Please try again."
            )
    elif isinstance(exception, requests.HTTPError):
        server_message = decode_sentinelhub_err_msg(exception.response)
        message += f'\nServer response: "{server_message}"'

    return message



"""Module defining custom package exceptions."""

from __future__ import annotations

import functools
import warnings
from typing import Any, Callable

import requests


class BaseSentinelHubException(Exception):
    """A base package exception"""


class DownloadFailedException(BaseSentinelHubException):
    """General exception which is raised whenever download fails"""

    def __init__(self, msg: str, *, request_exception: requests.RequestException | None = None) -> None:
        super().__init__(msg)
        self.request_exception = request_exception


class OutOfRequestsException(DownloadFailedException):
    """An exception raised whenever download cannot be done because user has run out of requests or processing units"""


class AwsDownloadFailedException(DownloadFailedException):
    """This exception is raised when download fails because of a missing file in AWS"""


class ImageDecodingError(BaseSentinelHubException):
    """This exception is raised when downloaded image is not properly decoded"""


class MissingDataInRequestException(BaseSentinelHubException):
    """This exception is raised when an iteration is performed on a request without data."""


class HashedNameCollisionException(BaseSentinelHubException):
    """This exception is raised when two different requests are assigned the same hashed name."""


class SHDeprecationWarning(DeprecationWarning):
    """A custom deprecation warning for sentinelhub-py package"""


class SHUserWarning(UserWarning):
    """A custom user warning for sentinelhub-py package"""


class SHRuntimeWarning(RuntimeWarning):
    """A custom runtime warning for sentinelhub-py package"""


class SHRateLimitWarning(SHRuntimeWarning):
    """A custom runtime warning in case user hit the rate limit for downloads"""


warnings.simplefilter("default", SHDeprecationWarning)
warnings.simplefilter("default", SHUserWarning)
warnings.simplefilter("always", SHRuntimeWarning)
warnings.simplefilter("always", SHRateLimitWarning)


# THIS SHOULD BE REMOVED IN THE FUTURE, after rest of libraries transition away.
def deprecated_function(
    category: type[DeprecationWarning] = SHDeprecationWarning, message_suffix: str | None = None
) -> Callable[[Callable], Callable]:
    """A parametrized function decorator, which signals that the function has been deprecated when called.

    Has to use paranthesis even when no custom parameters are used, e.g. `@deprecated_function()`.
    """

    def deco(func: Callable) -> Callable:
        message = f"Function `{func.__name__}` has been deprecated."
        if message_suffix:
            message += " " + message_suffix

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(message, category=category, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return deco


# THIS SHOULD BE REMOVED IN THE FUTURE, after rest of libraries transition away.
def deprecated_class(
    category: type[DeprecationWarning] = SHDeprecationWarning, message_suffix: str | None = None
) -> Callable[[type], type]:
    """A parametrized class decorator, which signals that the class has been deprecated when initialized.

    Has to use paranthesis even when no custom parameters are used, e.g. `@deprecated_class()`.
    """

    def deco(class_object: type) -> type:
        message = f"Class `{class_object.__name__}` has been deprecated."
        if message_suffix:
            message += " " + message_suffix

        old_init = class_object.__init__  # type: ignore[misc]

        def warn_and_init(self: Any, *args: Any, **kwargs: Any) -> None:
            warnings.warn(message, category=category, stacklevel=2)
            old_init(self, *args, **kwargs)

        class_object.__init__ = warn_and_init  # type: ignore[misc]
        return class_object

    return deco



"""
Module implementing some common utility functions
"""

# ruff: noqa: FA100
# do not use `from __future__ import annotations`, it clashes with `dataclass_json`
from enum import Enum
from typing import Any, Dict, Optional, Type, TypedDict

from dataclasses_json import LetterCase
from dataclasses_json import config as dataclass_config

from sentinelhub.types import JsonDict

from ..geometry import Geometry
from ..time_utils import parse_time, serialize_time

datetime_config = dataclass_config(
    encoder=lambda time: serialize_time(time, use_tz=True) if time else None,
    decoder=lambda time: parse_time(time, force_datetime=True) if time else None,
    letter_case=LetterCase.CAMEL,
)

geometry_config = dataclass_config(
    encoder=Geometry.get_geojson,
    decoder=lambda geojson: Geometry.from_geojson(geojson) if geojson else None,
    exclude=lambda geojson: geojson is None,
    letter_case=LetterCase.CAMEL,
)


def enum_config(enum_class: Type[Enum]) -> Dict[str, dict]:
    """Given an Enum class it provide an object for serialization/deserialization"""
    return dataclass_config(
        encoder=lambda enum_item: enum_item.value,
        decoder=lambda item: enum_class(item) if item else None,
        exclude=lambda item: item is None,
        letter_case=LetterCase.CAMEL,
    )


def _update_other_args(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> None:
    """Function for a recursive update of `dict1` with `dict2`. The function loops over the keys in `dict2` and
    only the non-dict like values are assigned to the specified keys.
    """
    for key, value in dict2.items():
        if isinstance(value, dict) and key in dict1:
            _update_other_args(dict1[key], value)
        else:
            dict1[key] = value


def remove_undefined(payload: dict) -> dict:
    """Takes a dictionary and removes keys without value"""
    return {name: value for name, value in payload.items() if value is not None}


class AccessSpecification(TypedDict):
    """Specification of a S3 input or output."""

    s3: JsonDict


def s3_specification(
    url: str,
    access_key: Optional[str] = None,
    secret_access_key: Optional[str] = None,
    iam_role_arn: Optional[str] = None,
    region: Optional[str] = None,
) -> AccessSpecification:
    """A helper method to build a dictionary used for specifying S3 paths. Consult the
    `access documentation <https://docs.sentinel-hub.com/api/latest/api/batch-statistical/#aws-bucket-access>`__
    for more information.

    In general either use `iam_role_arn` or `access_key` plus `secret_access_key`.

    :param url: A URL pointing to an S3 bucket or an object in an S3 bucket.
    :param access_key: AWS access key that allows programmatic access to the S3 bucket specified in the `url` field.
    :param secret_access_key: AWS secret access key which must correspond to the AWS access key.
    :param iam_role_arn: IAM role ARN, which allows programmatic access to the S3 bucket specified in the `url` field
        using the recommended assume IAM role flow.
    :param region: The region where the S3 bucket is located. If omitted, the region of the Sentinel Hub deployment
        that the request is submitted to is assumed.
    :return: A dictionary of S3 specifications used by the Batch Statistical API
    """
    if (iam_role_arn is None) == (access_key is None or secret_access_key is None):
        raise ValueError("Either specify `iam_role_arn` or both `access_key` and `secret_access_key`.")
    s3_access = {
        "url": url,
        "accessKey": access_key,
        "secretAccessKey": secret_access_key,
        "iamRoleARN": iam_role_arn,
        "region": region,
    }

    return {"s3": remove_undefined(s3_access)}



"""
Module implementing model classes to store download-related parameters and data.
"""

from __future__ import annotations

import datetime as dt
import functools
import hashlib
import json
import os
import platform
import warnings
from dataclasses import dataclass, field, fields
from typing import Any

from requests import Response

from ..constants import MimeType, RequestType
from ..decoding import decode_data
from ..exceptions import SHRuntimeWarning
from ..io_utils import read_data, write_data
from ..types import JsonDict


@dataclass
class DownloadRequest:
    """A class defining a single download request.

    The class is a container with all parameters needed to execute a single download request and save or return
    downloaded data.

    :param url: A URL from where to download
    :param headers: Headers of an HTTP request
    :param request_type: Type of request, either GET or POST. Default is `RequestType.GET`
    :param post_values: A dictionary of values that will be sent with a POST request. Default is `None`
    :param use_session: A flag that specifies if the download request will require a session header
    :param data_type: An expected file format of downloaded data. Default is `MimeType.RAW`
    :param save_response: A flag defining if the downloaded data will be saved to disk. Default is `False`.
    :param data_folder: A folder path where the fetched data will be (or already is) saved. Default is `None`
    :param filename: A custom filename where the data will be saved. By default, data will be saved in a folder
        which name are hashed request parameters.
    :param return_data: A flag defining if the downloaded data will be returned as an output of download procedure.
        Default is `True`.
    :param extra_params: Any additional parameters.
    """

    url: str | None = None
    headers: JsonDict = field(default_factory=dict)
    request_type: RequestType = RequestType.GET
    post_values: JsonDict | None = None
    use_session: bool = False
    data_type: MimeType = MimeType.RAW
    save_response: bool = False
    data_folder: str | None = None
    filename: str | None = None
    return_data: bool = True
    extra_params: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Additional parsing of init parameters."""
        self.request_type = RequestType(self.request_type)
        self.data_type = MimeType(self.data_type)

    def __hash__(self) -> int:
        """This dataclass is mutable, but we still assign its id as its hash."""
        return id(self)

    def raise_if_invalid(self) -> None:
        """Method that raises an error if something is wrong with request parameters

        :raises: ValueError
        """
        if self.save_response and self.data_folder is None:
            raise ValueError(
                "Data folder is not specified. Please give a data folder name in the initialization of your request."
            )

    def get_request_params(self, include_metadata: bool = False) -> JsonDict:
        """Provides parameters that define the request in form of a dictionary

        :param include_metadata: A flag defining if also metadata parameters should be included, such as headers and
            current time
        :return: A dictionary of parameters
        """
        params = {"url": self.url, "payload": self.post_values}
        if include_metadata:
            params = {**params, "headers": self.headers, "timestamp": dt.datetime.now().isoformat()}
        return params

    def get_hashed_name(self) -> str:
        """It takes request url and payload and calculates a unique hashed string from them.

        :return: A hashed string
        """
        params = self.get_request_params(include_metadata=False)
        hashable = json.dumps(params)

        return hashlib.md5(hashable.encode("utf-8")).hexdigest()

    def get_relative_paths(self) -> tuple[str | None, str]:
        """A method that calculates file paths relative to `data_folder`

        :return: Returns a pair of file paths, a request payload path and a response path. If request path is not
            defined it returns `None`.
        """

        if self.filename is not None:
            return None, self.filename

        hashed_name = self.get_hashed_name()

        request_path = os.path.join(hashed_name, "request.json")
        response_path = os.path.join(hashed_name, f"response.{self.data_type.extension}")

        return request_path, response_path

    def get_storage_paths(self) -> tuple[str | None, str | None]:
        """A method that calculates file paths where request payload and response will be saved.

        :return: Returns a pair of file paths, a request payload path and a response path. Each of them can also be
            `None` if it is not defined.
        """
        if self.data_folder is None:
            return None, None

        request_path, response_path = self.get_relative_paths()

        if request_path is not None:
            request_path = os.path.join(self.data_folder, request_path)
        response_path = os.path.join(self.data_folder, response_path)

        self._check_path(response_path)
        return request_path, response_path

    @staticmethod
    def _check_path(file_path: str) -> None:
        """Checks file path and warns about potential problems during saving"""
        message_problem = None
        if len(file_path) > 255 and platform.system() == "Windows":
            message_problem = "File path"
        elif len(os.path.basename(file_path)) > 255:
            message_problem = "Filename of"

        if message_problem:
            message = (
                f"{message_problem} {file_path} is longer than 255 character which might cause an error while "
                "saving on disk"
            )
            warnings.warn(message, category=SHRuntimeWarning)


@dataclass(frozen=True)
class DownloadResponse:
    """A class defining a single download response.

    :param request: A download request object for which the response is obtained.
    :param content: Raw encoded data of the response.
    :param headers: Headers obtained with the response.
    :param status_code: Status code of the response.
    :param elapsed: Number of seconds it took to obtain the response.
    """

    request: DownloadRequest
    content: bytes
    headers: JsonDict = field(default_factory=dict)
    status_code: int | None = None
    elapsed: float | None = None

    @classmethod
    def from_response(cls, response: Response, request: DownloadRequest) -> DownloadResponse:
        """Creates `DownloadResponse` object from obtained from a service response and request info.

        :param: A service response object.
        :param: A request for which response was obtained.
        :return: An instance of a download response object.
        """
        return cls(
            request=request,
            content=response.content,
            headers=dict(response.headers),
            status_code=response.status_code,
            elapsed=response.elapsed.total_seconds(),
        )

    @classmethod
    def from_local(cls, request: DownloadRequest) -> DownloadResponse:
        """Creates `DownloadResponse` object by loading it from locally cached data.

        :param request: A request object for which data is cached locally.
        :return: An instance of a download response object.
        """
        request_path, response_path = request.get_storage_paths()
        if response_path is None:
            raise ValueError("Cannot load cached data because response path isn't defined")

        content = read_data(response_path, data_format=MimeType.RAW)

        response_builder = functools.partial(cls, request=request, content=content)
        if request_path is None:
            return response_builder()

        response_info = read_data(request_path, data_format=MimeType.JSON).get("response")
        if response_info is None:
            return response_builder()

        return response_builder(
            headers=response_info.get("headers", {}),
            status_code=response_info.get("status_code"),
            elapsed=response_info.get("elapsed"),
        )

    def to_local(self) -> None:
        """Caches data about a request and a response locally."""
        request_path, response_path = self.request.get_storage_paths()
        if response_path is None:
            raise ValueError("Cannot cache data because response path isn't defined")

        write_data(response_path, self.content, data_format=MimeType.RAW)

        if request_path is None:
            return

        info = {
            "request": self.request.get_request_params(include_metadata=True),
            "response": {
                "headers": self.headers,
                "status_code": self.status_code,
                "elapsed": self.elapsed,
            },
        }
        write_data(request_path, info, data_format=MimeType.JSON)

    @property
    def response_type(self) -> MimeType:
        """Provides the expected mime type of the response data."""
        if self.request.data_type is not MimeType.RAW:
            return self.request.data_type

        content_type = self.headers.get("Content-Type") or self.headers.get("content-type")
        if content_type:
            return MimeType.from_string(content_type)
        return MimeType.RAW

    def decode(self) -> Any:
        """Decodes binary data into a Python object."""
        return decode_data(self.content, data_type=self.response_type)

    def derive(self, **params: Any) -> DownloadResponse:
        """Create a new response by changing some parameters of the existing one.

        :param params: Any of `DownloadResponse` attributes.
        :return: A new instance of `DownloadResponse` with modified parameters
        """
        derived_params = {_field.name: getattr(self, _field.name) for _field in fields(self)}
        derived_params.update(params)

        return DownloadResponse(**derived_params)



"""
Implementation of base Sentinel Hub interfaces
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any

from ..base import DataRequest
from ..constants import MimeType, MosaickingOrder, RequestType, ResamplingType
from ..data_collections import DataCollection, OrbitDirection
from ..download import DownloadRequest
from ..geometry import BBox, Geometry
from ..time_utils import RawTimeIntervalType, parse_time_interval, serialize_time
from .utils import _update_other_args


class SentinelHubBaseApiRequest(DataRequest, metaclass=ABCMeta):
    """A base class for Sentinel Hub interfaces"""

    _SERVICE_ENDPOINT = ""
    payload: dict[str, Any] = {}  # noqa: RUF012

    @property
    @abstractmethod
    def mime_type(self) -> MimeType:
        """The mime type of the request."""

    def create_request(self) -> None:
        """Prepares a download request"""
        headers = {"content-type": MimeType.JSON.get_string(), "accept": self.mime_type.get_string()}
        base_url = self._get_base_url()
        self.download_list = [
            DownloadRequest(
                request_type=RequestType.POST,
                url=f"{base_url}/api/v1/{self._SERVICE_ENDPOINT}",
                post_values=self.payload,
                data_folder=self.data_folder,
                save_response=bool(self.data_folder),
                data_type=self.mime_type,
                headers=headers,
                use_session=True,
            )
        ]

    @staticmethod
    def input_data(
        data_collection: DataCollection,
        *,
        identifier: str | None = None,
        time_interval: RawTimeIntervalType | None = None,
        maxcc: float | None = None,
        mosaicking_order: MosaickingOrder | None = None,
        upsampling: ResamplingType | None = None,
        downsampling: ResamplingType | None = None,
        other_args: dict[str, Any] | None = None,
    ) -> InputDataDict:
        """Generate the `input data` part of the request body

        :param data_collection: One of supported Process API data collections.
        :param identifier: A collection identifier that can be referred to in the evalscript. Parameter is referenced
            as `"id"` in service documentation. To learn more check
            `data fusion documentation <https://docs.sentinel-hub.com/api/latest/data/data-fusion>`__.
        :param time_interval: A time interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD or
            a datetime object
        :param maxcc: Maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is 1.0.
        :param mosaicking_order: Mosaicking order, which has to be either 'mostRecent', 'leastRecent' or 'leastCC'.
        :param upsampling: A type of upsampling to apply on data
        :param downsampling: A type of downsampling to apply on data
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :return: A dictionary-like object that also contains additional attributes
        """
        input_data_dict: dict[str, Any] = {
            "type": data_collection.api_id,
        }
        if identifier:
            input_data_dict["id"] = identifier

        data_filters = _get_data_filters(data_collection, time_interval, maxcc, mosaicking_order)
        if data_filters:
            input_data_dict["dataFilter"] = data_filters

        processing_params = _get_processing_params(upsampling, downsampling)
        if processing_params:
            input_data_dict["processing"] = processing_params

        if other_args:
            _update_other_args(input_data_dict, other_args)

        return InputDataDict(input_data_dict, service_url=data_collection.service_url)

    @staticmethod
    def bounds(
        bbox: BBox | None = None, geometry: Geometry | None = None, other_args: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """Generate a `bound` part of the API request

        :param bbox: Bounding box describing the area of interest.
        :param geometry: Geometry describing the area of interest.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        if bbox is None and geometry is None:
            raise ValueError("'bbox' and/or 'geometry' have to be provided.")

        if bbox and not isinstance(bbox, BBox):
            raise ValueError("'bbox' should be an instance of sentinelhub.BBox")

        if geometry and not isinstance(geometry, Geometry):
            raise ValueError("'geometry' should be an instance of sentinelhub.Geometry")

        if bbox and geometry and bbox.crs != geometry.crs:
            raise ValueError("bbox and geometry should be in the same CRS")

        crs = bbox.crs if bbox else geometry.crs  # type: ignore[union-attr]

        request_bounds: dict[str, Any] = {"properties": {"crs": crs.opengis_string}}

        if bbox:
            request_bounds["bbox"] = list(bbox)

        if geometry:
            request_bounds["geometry"] = geometry.get_geojson(with_crs=False)

        if other_args:
            _update_other_args(request_bounds, other_args)

        return request_bounds

    def _get_base_url(self) -> str:
        """It decides which base URL to use. Restrictions from data collection definitions overrule the
        settings from config object. In case different collections have different restrictions then
        `SHConfig.sh_base_url` breaks the tie in case it matches one of the data collection URLs.
        """
        data_collection_urls = tuple(
            {
                input_data_dict.service_url.rstrip("/")
                for input_data_dict in self.payload["input"]["data"]
                if isinstance(input_data_dict, InputDataDict) and input_data_dict.service_url is not None
            }
        )
        config_base_url = self.config.sh_base_url.rstrip("/")

        if not data_collection_urls:
            return config_base_url

        if len(data_collection_urls) == 1:
            return data_collection_urls[0]

        if config_base_url in data_collection_urls:
            return config_base_url

        raise ValueError(
            f"Given data collections are restricted to different services: {data_collection_urls}\n"
            "Configuration parameter sh_base_url cannot break the tie because it is set to a different"
            f"service: {config_base_url}"
        )


class InputDataDict(dict):
    """An input data dictionary which also holds additional attributes"""

    def __init__(self, input_data_dict: dict[str, Any], *, service_url: str | None = None):
        """
        :param input_data_dict: A normal dictionary with input parameters
        :param service_url: A service URL defined by a data collection
        """
        super().__init__(input_data_dict)
        self.service_url = service_url

    def __repr__(self) -> str:
        """Modified dictionary representation that also shows additional attributes"""
        normal_dict_repr = super().__repr__()
        return f"{self.__class__.__name__}({normal_dict_repr}, service_url={self.service_url})"


def _get_data_filters(
    data_collection: DataCollection,
    time_interval: RawTimeIntervalType | None,
    maxcc: float | None,
    mosaicking_order: MosaickingOrder | None,
) -> dict[str, Any]:
    """Builds a dictionary of data filters for Process API"""
    data_filter: dict[str, Any] = {}

    if time_interval:
        start_time, end_time = serialize_time(parse_time_interval(time_interval, allow_undefined=True), use_tz=True)
        data_filter["timeRange"] = {"from": start_time, "to": end_time}

    if maxcc is not None:
        if maxcc < 0 or maxcc > 1:
            raise ValueError("maxcc should be a float on an interval [0, 1]")

        data_filter["maxCloudCoverage"] = int(maxcc * 100)

    if mosaicking_order:
        data_filter["mosaickingOrder"] = MosaickingOrder(mosaicking_order).value

    return {**data_filter, **_get_data_collection_filters(data_collection)}


def _get_data_collection_filters(data_collection: DataCollection) -> dict[str, Any]:
    """Builds a dictionary of filters for Process API from a data collection definition"""
    filters: dict[str, Any] = {}

    if data_collection.swath_mode:
        filters["acquisitionMode"] = data_collection.swath_mode.upper()

    if data_collection.polarization:
        filters["polarization"] = data_collection.polarization.upper()

    if data_collection.resolution:
        filters["resolution"] = data_collection.resolution.upper()

    if data_collection.orbit_direction and data_collection.orbit_direction.upper() != OrbitDirection.BOTH:
        filters["orbitDirection"] = data_collection.orbit_direction.upper()

    if data_collection.timeliness:
        filters["timeliness"] = data_collection.timeliness

    if data_collection.dem_instance:
        filters["demInstance"] = data_collection.dem_instance

    return filters


def _get_processing_params(upsampling: ResamplingType | None, downsampling: ResamplingType | None) -> dict[str, Any]:
    """Builds a dictionary of processing parameters for Process API"""
    processing_params: dict[str, Any] = {}

    if upsampling:
        processing_params["upsampling"] = ResamplingType(upsampling).value

    if downsampling:
        processing_params["downsampling"] = ResamplingType(downsampling).value

    return processing_params



"""
Deprecated module for Batch Processing v2 API. The contents have been moved to `sentinelhub.api.batch.process`.
"""

import warnings

from sentinelhub.api.batch.process import *  # noqa: F403 # pylint: disable=unused-wildcard-import, wildcard-import
from sentinelhub.exceptions import SHDeprecationWarning

warnings.warn(
    "The module sentinelhub.api.batch.process_v2 has been renamed, please use sentinelhub.api.batch.process instead.",
    SHDeprecationWarning,
    stacklevel=2,
)



"""
Module defining evalscript generation utilities
"""

from __future__ import annotations

import numpy as np

from .data_collections import DataCollection
from .data_collections_bands import Band, Unit

DTYPE_TO_SAMPLE_TYPE: dict[type, str] = {
    bool: "UINT8",
    np.uint8: "UINT8",
    np.uint16: "UINT16",
    np.float32: "FLOAT32",
}

EVALSCRIPT_TEMPLATE = """
//VERSION=3

function setup() {{
    return {{
        input: [{{
            bands: [{input_names}],
            units: [{input_units}]
        }}],
        output: [{output_spec}]
    }}
}}

function updateOutputMetadata(scenes, inputMetadata, outputMetadata) {{
    outputMetadata.userData = {{
        "norm_factor":  inputMetadata.normalizationFactor
    }}
}}

function evaluatePixel(sample) {{
    return {{ {return_spec} }};
}}
"""


def parse_data_collection_bands(data_collection: DataCollection, bands: list[str]) -> list[Band]:
    """Checks that all requested bands are available and returns the band information for further processing

    :param data_collection: A collection of requested satellite data.
    :param bands: A list of band or meta band names to use in the evalscript.
    """

    requested_bands = []
    band_info_dict = {band_info.name: band_info for band_info in data_collection.bands + data_collection.metabands}

    for band_name in bands:
        if band_name not in band_info_dict:
            raise ValueError(
                f"Data collection {data_collection} does not have specifications for band {band_name}.\n"
                f"Available bands are:\n{[band.name for band in data_collection.bands]}\nand meta-bands:\n"
                f"{[band.name for band in data_collection.metabands]}."
            )
        requested_bands.append(band_info_dict[band_name])

    return requested_bands


def generate_evalscript(
    data_collection: DataCollection,
    bands: list[str] | None = None,
    meta_bands: list[str] | None = None,
    merged_bands_output: str | None = None,
    prioritize_dn: bool = True,
) -> str:
    """Generate an evalscript based on the provided specifications. This utility supports generating only evalscripts
    with the mosaicking option set to `SIMPLE`.

    :param data_collection: A collection of requested satellite data.
    :param bands: A list of band names to use in the evalscript. Defaults to using all bands provided by the collection.
    :param meta_bands: A list of meta band names to use in the evalscript. By default no meta bands are added.
    :param merged_bands_output: If provided, bands will be concatenated into a single multi-band tiff with this name.
    :param prioritize_dn: Use DN units if possible. Default is True. If DN units are not available, the default units
        for each specific band are used. DN units will be used regardless of the flag if they are the only possible
        choice.
    """

    band_names = bands if bands is not None else [band.name for band in data_collection.bands]
    meta_band_names = meta_bands if meta_bands is not None else []

    input_names, input_units = [], []
    sample_type_map: dict[str, str] = {}
    requested_bands = parse_data_collection_bands(data_collection, band_names + meta_band_names)

    for band in requested_bands:
        unit_choice_idx = band.units.index(Unit.DN) if (prioritize_dn and Unit.DN in band.units) else 0
        sample_type_map[band.name] = DTYPE_TO_SAMPLE_TYPE[band.output_types[unit_choice_idx]]
        input_names.append(f'"{band.name}"')
        input_units.append(f'"{band.units[unit_choice_idx].value}"')

    output_spec, return_spec = [], []
    if merged_bands_output is not None:
        sample_type = sample_type_map[band_names[0]]
        output_spec.append(f'{{id: "{merged_bands_output}", bands: {len(band_names)}, sampleType: "{sample_type}"}}')
        return_spec.append(f"{merged_bands_output}: [{', '.join(f'sample.{band_name}' for band_name in band_names)}]")

    for band_name, sample_type in sample_type_map.items():
        # potentially skip bands if merged bands output specified, but still run for meta bands
        if merged_bands_output is not None and band_name in band_names:
            continue

        output_spec.append(f'{{id: "{band_name}", bands: 1, sampleType: "{sample_type}"}}')
        return_spec.append(f"{band_name}: [sample.{band_name}]")

    return EVALSCRIPT_TEMPLATE.format(
        input_names=", ".join(input_names),
        input_units=", ".join(input_units),
        output_spec=", ".join(output_spec),
        return_spec=", ".join(return_spec),
    )



"""
Utility tools for writing unit tests for packages which rely on `sentinelhub-py`
"""

from __future__ import annotations

import os
from typing import Any, Callable

import numpy as np
import pytest


def get_input_folder(current_file: str) -> str:
    """Use fixtures if possible. This is meant only for test cases"""
    return os.path.join(os.path.dirname(os.path.realpath(current_file)), "TestInputs")


def get_output_folder(current_file: str) -> str:
    """Use fixtures if possible. This is meant only for test cases"""
    return os.path.join(os.path.dirname(os.path.realpath(current_file)), "TestOutputs")


def assert_statistics_match(
    data: np.ndarray,
    exp_shape: tuple[int, ...] | None = None,
    exp_dtype: None | type | np.dtype = None,
    exp_min: float | None = None,
    exp_max: float | None = None,
    exp_mean: float | None = None,
    exp_median: float | None = None,
    exp_std: float | None = None,
    rel_delta: float | None = None,
    abs_delta: float | None = None,
) -> None:
    """Validates basic statistics of data array
    :param data: Data array
    :param exp_shape: Expected shape
    :param exp_dtype: Expected dtype
    :param exp_min: Expected minimal value
    :param exp_max: Expected maximal value
    :param exp_mean: Expected mean value
    :param exp_median: Expected median value
    :param exp_std: Expected standard deviation value
    :param rel_delta: Precision of validation (relative)
    :param abs_delta: Precision of validation (absolute)
    """

    stats_suite: dict[str, tuple[Callable[[np.ndarray], Any], Any]] = {
        "shape": (lambda array: array.shape, exp_shape),
        "dtype": (lambda array: array.dtype, exp_dtype),
        "min": (np.nanmin, exp_min),
        "max": (np.nanmax, exp_max),
        "mean": (np.nanmean, exp_mean),
        "median": (np.nanmedian, exp_median),
        "std": (np.nanstd, exp_std),
    }

    is_precise = {"shape", "dtype"}

    data_stats: dict[str, Any] = {}
    exp_stats: dict[str, Any] = {}
    for name, (func, expected) in stats_suite.items():
        if expected is not None:
            data_stats[name] = func(data)
            exp_stats[name] = expected if name in is_precise else pytest.approx(expected, rel=rel_delta, abs=abs_delta)

    assert data_stats == exp_stats, "Statistics differ from expected values"



"""
Download process for Sentinel Hub Statistical API
"""

from __future__ import annotations

import concurrent.futures
import copy
import json
import logging
from typing import Any

from ..exceptions import DownloadFailedException
from ..types import JsonDict
from .models import DownloadRequest, DownloadResponse
from .sentinelhub_client import SentinelHubDownloadClient

LOGGER = logging.getLogger(__name__)


class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
    """A special download client for Sentinel Hub Statistical API

    Beside a normal download from Sentinel Hub services it implements an additional process of retrying and caching.
    """

    _RETRIABLE_ERRORS = ("EXECUTION_ERROR", "TIMEOUT")

    def __init__(self, *args: Any, n_interval_retries: int = 1, max_retry_threads: int = 5, **kwargs: Any):
        """
        :param n_interval_retries: Number of retries if a request fails just for a certain timestamp. (This parameter
            is experimental and might be changed in the future.)
        :param max_retry_threads: Number of threads used for retrying. (This parameter is experimental and might be
            changed in the future.)
        """
        super().__init__(*args, **kwargs)

        self.n_interval_retries = n_interval_retries
        self.max_retry_threads = max_retry_threads

    def _process_response(self, request: DownloadRequest, response: DownloadResponse) -> DownloadResponse:
        """After downloading the response for all timestamps this method handles redownload for those timestamps for
        which download failed."""
        stats_response = response.decode()

        failed_time_intervals: dict[int, Any] = {}
        for index, stat_info in enumerate(stats_response["data"]):
            if self._has_retriable_error(stat_info):
                failed_time_intervals[index] = stat_info["interval"]

        n_succeeded_intervals = 0
        if failed_time_intervals:
            LOGGER.debug("Failed for %s intervals, retrying by downloading per interval", len(failed_time_intervals))
            retried_responses = self._download_per_interval(request, failed_time_intervals)
            n_succeeded_intervals = sum("error" not in stat_info for stat_info in retried_responses.values())

            stats_response["data"] = [
                retried_responses.get(index, stat_info) for index, stat_info in enumerate(stats_response["data"])
            ]

        if n_succeeded_intervals == 0:
            return response

        new_content = json.dumps(stats_response).encode("utf-8")
        return response.derive(content=new_content)

    def _download_per_interval(self, request: DownloadRequest, time_intervals: dict[int, Any]) -> dict:
        """Download statistics per each time interval"""
        interval_requests = []
        for time_interval in time_intervals.values():
            interval_request = copy.deepcopy(request)
            if interval_request.post_values is None or "aggregation" not in interval_request.post_values:
                raise ValueError("Unable to configure request for retrying by interval.")

            interval_request.post_values["aggregation"]["timeRange"] = time_interval
            interval_requests.append(interval_request)

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_retry_threads) as executor:
            stat_info_responses = list(executor.map(self._execute_single_stat_download, interval_requests))

        return dict(zip(time_intervals, stat_info_responses))

    def _execute_single_stat_download(self, request: DownloadRequest) -> JsonDict:
        """Makes sure a download for a single time interval is retried"""
        for retry_count in range(self.n_interval_retries):
            response = self._execute_download(request)
            stat_response = response.decode()
            stat_info = stat_response["data"][0]

            if not self._has_retriable_error(stat_info) or retry_count == self.n_interval_retries - 1:
                return stat_info

        raise DownloadFailedException("No more interval retries available, download unsuccessful")

    def _has_retriable_error(self, stat_info: JsonDict) -> bool:
        """Checks if a dictionary of Stat API info for a single time interval has an error that can fixed by retrying
        a request
        """
        error_type = stat_info.get("error", {}).get("type")
        return error_type in self._RETRIABLE_ERRORS



"""
Utility functions to read/write image data from/to file
"""

from __future__ import annotations

import csv
import json
import logging
import os
from typing import IO, Any, Callable, Literal
from xml.etree import ElementTree

import numpy as np
import tifffile as tiff
from PIL import Image

from .constants import MimeType
from .decoding import decode_image_with_pillow, decode_jp2_image, decode_tar, get_data_format

LOGGER = logging.getLogger(__name__)

CSV_DELIMITER = ";"


def read_data(filename: str, data_format: MimeType | None = None) -> Any:
    """Read image data from file

    This function reads input data from file. The format of the file
    can be specified in ``data_format``. If not specified, the format is
    guessed from the extension of the filename.

    :param filename: filename to read data from
    :param data_format: format of filename. Default is `None`
    :return: data read from filename
    :raises: exception if filename does not exist
    """
    if not os.path.exists(filename):
        raise FileNotFoundError(f"Filename {filename} does not exist")

    if not isinstance(data_format, MimeType):
        data_format = get_data_format(filename)

    reader = _get_reader(data_format)

    try:
        return reader(filename)
    except BaseException as exception:
        # In case a procedure would read a lot of files and one would be corrupt this helps us figure out which one
        LOGGER.debug("Failed to read from file: %s", filename)
        raise exception


def _get_reader(data_format: MimeType) -> Callable[[str], Any]:
    """Provides a function for reading data in a given data format"""
    if data_format is MimeType.TIFF:
        return tiff.imread
    if data_format is MimeType.JP2:
        return _open_file_and_read(decode_jp2_image, "rb")
    if data_format.is_image_format():
        return decode_image_with_pillow

    available_readers: dict[MimeType, Callable[[str], Any]] = {
        MimeType.TAR: _open_file_and_read(decode_tar, "rb"),  # type: ignore[arg-type]
        MimeType.TXT: _open_file_and_read(lambda file: file.read(), "r"),
        MimeType.RAW: _open_file_and_read(lambda file: file.read(), "rb"),
        MimeType.CSV: _read_csv,
        MimeType.JSON: _open_file_and_read(json.load, "rb"),
        MimeType.XML: ElementTree.parse,
        MimeType.GML: ElementTree.parse,
        MimeType.SAFE: ElementTree.parse,
        MimeType.NPY: np.load,
    }

    if data_format not in available_readers:
        raise ValueError(f"Reading data format {data_format} is not supported.")

    return available_readers[data_format]


def _open_file_and_read(reader: Callable[[IO], Any], mode: Literal["r", "rb"]) -> Callable[[str], Any]:
    def new_reader(filename: str) -> Any:
        with open(filename, mode) as file:
            return reader(file)

    return new_reader


def _read_csv(filename: str, delimiter: str = CSV_DELIMITER) -> list:
    """Read data from CSV file

    :param filename: name of CSV file to be read
    :param delimiter: type of CSV delimiter. Default is ``;``
    :return: data stored in CSV file as list
    """
    with open(filename) as file:
        return list(csv.reader(file, delimiter=delimiter))


def write_data(  # noqa: C901
    filename: str, data: Any, data_format: MimeType | None = None, compress: bool = False, add: bool = False
) -> None:
    """Write image data to file

    Function to write image data to specified file. If file format is not provided
    explicitly, it is guessed from the filename extension. If format is TIFF, geo
    information and compression can be optionally added.

    :param filename: name of file to write data to
    :param data: image data to write to file
    :param data_format: format of output file. Default is `None`
    :param compress: Compress data. Default is `False`
    :param add: Append to existing file. Only supported for TXT. Default is `False`
    :raises: exception if numpy format is not supported or file cannot be written
    """
    _create_parent_folder(filename)

    if not isinstance(data_format, MimeType):
        data_format = get_data_format(filename)

    if data_format is MimeType.TIFF:
        tiff.imwrite(filename, data, compression=("lzma" if compress else None))

    elif data_format.is_image_format():
        Image.fromarray(data).save(filename)

    elif data_format is MimeType.NPY:
        np.save(filename, data)

    elif data_format in (MimeType.XML, MimeType.GML):
        data.write(filename)

    elif data_format is MimeType.TXT:
        with open(filename, "a" if add else "w") as file:
            print(data, end="", file=file)

    elif data_format is MimeType.RAW:
        with open(filename, "wb") as file:
            file.write(data)

    elif data_format is MimeType.CSV:
        with open(filename, "w") as file:
            csv_writer = csv.writer(file, delimiter=CSV_DELIMITER)
            for line in data:
                csv_writer.writerow(line)

    elif data_format is MimeType.JSON:
        with open(filename, "w") as file:
            json.dump(data, file, indent=4, sort_keys=True)

    else:
        raise ValueError(f"Writing data format {data_format} is not supported")


def _create_parent_folder(filename: str) -> None:
    path = os.path.dirname(filename)
    if path != "":
        os.makedirs(path, exist_ok=True)



"""
Module implementing an interface with
`Sentinel Hub Batch Processing API <https://docs.sentinel-hub.com/api/latest/api/batch-statistical/>`__.
"""

# ruff: noqa: FA100
# do not use `from __future__ import annotations`, it clashes with `dataclass_json`
import datetime as dt
import logging
from dataclasses import dataclass, field
from typing import Any, Optional, Sequence, Union

from dataclasses_json import CatchAll, LetterCase, Undefined, dataclass_json
from dataclasses_json import config as dataclass_config
from typing_extensions import deprecated

from ...exceptions import SHDeprecationWarning
from ...types import Json, JsonDict
from ..base_request import InputDataDict
from ..statistical import SentinelHubStatistical
from ..utils import AccessSpecification, datetime_config, enum_config, remove_undefined, s3_specification
from .base import BaseBatchClient, BaseBatchRequest, BatchRequestStatus, BatchUserAction

LOGGER = logging.getLogger(__name__)

BatchStatisticalRequestType = Union[str, dict, "BatchStatisticalRequest"]


class SentinelHubBatchStatistical(BaseBatchClient["BatchStatisticalRequest"]):
    """An interface class for Sentinel Hub Batch Statistical API

    Check `Batch Statistical API <https://docs.sentinel-hub.com/api/latest/reference/#tag/batch_statistical>`__ for more
    information.
    """

    s3_specification = staticmethod(s3_specification)

    @staticmethod
    def _get_service_url(base_url: str) -> str:
        """Provides URL to Batch Statistical API"""
        return f"{base_url}/api/v1/statistics/batch"

    def create(
        self,
        *,
        input_features: AccessSpecification,
        input_data: Sequence[Union[JsonDict, InputDataDict]],
        aggregation: JsonDict,
        calculations: Optional[JsonDict],
        output: AccessSpecification,
        **kwargs: Any,
    ) -> "BatchStatisticalRequest":
        """Create a new batch statistical request

        `Batch Statistical API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/createNewBatchStatisticsRequest>`__
        """
        payload = {
            "input": {"features": input_features, "data": list(input_data)},
            "aggregation": aggregation,
            "calculations": calculations,
            "output": output,
            **kwargs,
        }
        payload = remove_undefined(payload)

        url = self.service_url
        request_info = self.client.get_json_dict(url, post_values=payload, use_session=True)

        return BatchStatisticalRequest.from_dict(request_info)

    def create_from_request(
        self,
        statistical_request: SentinelHubStatistical,
        input_features: AccessSpecification,
        output: AccessSpecification,
        **kwargs: Any,
    ) -> "BatchStatisticalRequest":
        """Create a new batch statistical request from an existing statistical request.

        :param statistical_request: A Sentinel Hub Statistical request.
        :param input_features: A dictionary describing the S3 path and credentials to access the input GeoPackage.
        :param output: A dictionary describing the S3 path and credentials to access the output folder.
        :param kwargs: Any other arguments to be added to a dictionary of parameters
        :returns: A Batch Statistical request with the same calculations and aggregations but using geometries
            specified in the input GeoPackage.
        """

        return self.create(
            input_features=input_features,
            input_data=statistical_request.payload["input"]["data"],
            aggregation=statistical_request.payload["aggregation"],
            calculations=statistical_request.payload["calculations"],
            output=output,
            **kwargs,
        )

    def get_request(self, batch_request: BatchStatisticalRequestType) -> "BatchStatisticalRequest":
        """Collects information about a single batch request

        `Batch Statistical API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getSingleBatchStatisticalRequestById>`__

        :return: Batch request info
        """
        request_id = self._parse_request_id(batch_request)
        request_info = self.client.get_json_dict(url=self._get_processing_url(request_id), use_session=True)
        return BatchStatisticalRequest.from_dict(request_info)

    def get_status(self, batch_request: BatchStatisticalRequestType) -> JsonDict:
        """Collects information about a status of a request

        `Batch Statistical API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchStatisticalGetStatus>`__

        :return: Batch request status dictionary
        """
        request_id = self._parse_request_id(batch_request)
        endpoint_url = f"{self._get_processing_url(request_id)}/status"
        return self.client.get_json_dict(url=endpoint_url, use_session=True)

    def start_analysis(self, batch_request: BatchStatisticalRequestType) -> Json:
        """Starts analysis of a batch job request

        `Batch Statistical API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchStatisticalAnalyse>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "analyse")

    def start_job(self, batch_request: BatchStatisticalRequestType) -> Json:
        """Starts running a batch job

        `Batch Statistical API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchStartStatisticalRequest>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "start")

    @deprecated("The method `cancel_job` has been replaced with use `stop_job`.", category=SHDeprecationWarning)
    def cancel_job(self, batch_request: BatchStatisticalRequestType) -> Json:
        """Cancels a batch job

        `Batch Statistical API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchCancelStatisticalRequest>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "cancel")

    def stop_job(self, batch_request: BatchStatisticalRequestType) -> Json:
        """Stop a batch job

        `Batch Statistical API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#tag/batch_statistical/operation/batchStopStatisticalRequest>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "stop")


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass(repr=False)
class BatchStatisticalRequest(BaseBatchRequest):  # pylint: disable=abstract-method
    """A dataclass object that holds information about a batch statistical request"""

    # dataclass_json doesn't handle parameter inheritance correctly
    # pylint: disable=duplicate-code

    request_id: str = field(metadata=dataclass_config(field_name="id"))
    completion_percentage: float
    request: dict
    status: BatchRequestStatus = field(metadata=enum_config(BatchRequestStatus))
    user_id: Optional[str] = None
    created: Optional[dt.datetime] = field(metadata=datetime_config, default=None)
    cost_pu: Optional[float] = field(metadata=dataclass_config(field_name="costPU"), default=None)
    user_action: Optional[BatchUserAction] = field(metadata=enum_config(BatchUserAction), default=None)
    user_action_updated: Optional[str] = field(metadata=datetime_config, default=None)
    error: Optional[str] = None
    other_data: CatchAll = field(default_factory=dict)

    _REPR_PARAM_NAMES = (
        "request_id",
        "created",
        "status",
        "user_action",
        "cost_pu",
        "error",
        "completion_percentage",
    )



"""
Module for manipulation of geographical information
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, Tuple, cast

from typing_extensions import deprecated

from .constants import CRS
from .exceptions import SHDeprecationWarning

if TYPE_CHECKING:
    from .geometry import BBox

ERR = 0.1


def bbox_to_dimensions(bbox: BBox, resolution: float | tuple[float, float]) -> tuple[int, int]:
    """Calculates width and height in pixels for a given bbox of a given pixel resolution (in meters). The result is
    rounded to the nearest integers.

    :param bbox: bounding box
    :param resolution: Resolution of desired image in meters. It can be a single number or a tuple of two numbers -
        resolution in horizontal and resolution in vertical direction.
    :return: width and height in pixels for given bounding box and pixel resolution
    """
    utm_bbox = to_utm_bbox(bbox)
    east1, north1 = utm_bbox.lower_left
    east2, north2 = utm_bbox.upper_right

    resx, resy = resolution if isinstance(resolution, tuple) else (resolution, resolution)

    return round(abs(east2 - east1) / resx), round(abs(north2 - north1) / resy)


def bbox_to_resolution(bbox: BBox, width: int, height: int, meters: bool = True) -> tuple[float, float]:
    """Calculates pixel resolution for a given bbox of a given width and height. By default, it returns result in
    meters.

    :param bbox: bounding box
    :param width: width of bounding box in pixels
    :param height: height of bounding box in pixels
    :param meters: If `True` result will be given in meters, otherwise it will be given in units of current CRS
    :return: resolution east-west at north and south, and resolution north-south for given CRS
    :raises: ValueError if CRS is not supported
    """
    if meters:
        bbox = to_utm_bbox(bbox)
    east1, north1 = bbox.lower_left
    east2, north2 = bbox.upper_right
    return abs(east2 - east1) / width, abs(north2 - north1) / height


def get_image_dimension(bbox: BBox, width: int | None = None, height: int | None = None) -> int:
    """Given bounding box and one of the parameters width or height it will return the other parameter that will best
    fit the bounding box dimensions

    :param bbox: bounding box
    :param width: image width or `None` if height is unknown
    :param height: image height or `None` if height is unknown
    :return: width or height rounded to integer
    """
    utm_bbox = to_utm_bbox(bbox)
    east1, north1 = utm_bbox.lower_left
    east2, north2 = utm_bbox.upper_right
    if isinstance(width, int):
        return round(width * abs(north2 - north1) / abs(east2 - east1))
    if isinstance(height, int):
        return round(height * abs(east2 - east1) / abs(north2 - north1))
    raise ValueError("At least one of the parameters `width` and `height` must be given.")


def to_utm_bbox(bbox: BBox) -> BBox:
    """Transform bbox into UTM CRS

    :param bbox: bounding box
    :return: bounding box in UTM CRS
    """
    if CRS.is_utm(bbox.crs):
        return bbox
    lng, lat = bbox.middle
    utm_crs = get_utm_crs(lng, lat, source_crs=bbox.crs)
    return bbox.transform(utm_crs)


@deprecated("The function `get_utm_bbox` has been deprecated.", category=SHDeprecationWarning)
def get_utm_bbox(img_bbox: Sequence[float], transform: Sequence[float]) -> list[float]:
    """Get UTM coordinates given a bounding box in pixels and a transform

    :param img_bbox: boundaries of bounding box in pixels as `[row1, col1, row2, col2]`
    :param transform: georeferencing transform of the image, e.g. `(x_upper_left, res_x, 0, y_upper_left, 0, -res_y)`
    :return: UTM coordinates as [east1, north1, east2, north2]
    """
    east1, north1 = pixel_to_utm(img_bbox[0], img_bbox[1], transform)
    east2, north2 = pixel_to_utm(img_bbox[2], img_bbox[3], transform)
    return [east1, north1, east2, north2]


@deprecated(
    "The function `wgs84_to_utm` has been deprecated. Use `transform_point` and `get_utm_crs` instead.",
    category=SHDeprecationWarning,
)
def wgs84_to_utm(lng: float, lat: float, utm_crs: CRS | None = None) -> tuple[float, float]:
    """Convert WGS84 coordinates to UTM. If UTM CRS is not set it will be calculated automatically.

    :param lng: longitude in WGS84 system
    :param lat: latitude in WGS84 system
    :param utm_crs: UTM coordinate reference system enum constants
    :return: east, north coordinates in UTM system
    """
    if utm_crs is None:
        utm_crs = get_utm_crs(lng, lat)
    return transform_point((lng, lat), CRS.WGS84, utm_crs)


@deprecated(
    "The function `to_wgs84` has been deprecated. Use `transform_point` instead.", category=SHDeprecationWarning
)
def to_wgs84(east: float, north: float, crs: CRS) -> tuple[float, float]:
    """Convert any CRS with (east, north) coordinates to WGS84

    :param east: east coordinate
    :param north: north coordinate
    :param crs: CRS enum constants
    :return: latitude and longitude coordinates in WGS84 system
    """
    return transform_point((east, north), crs, CRS.WGS84)


def utm_to_pixel(
    east: float, north: float, transform: Sequence[float], truncate: bool = True
) -> tuple[float, float] | tuple[int, int]:
    """Convert a UTM coordinate to image coordinate given a transform

    :param east: east coordinate of point
    :param north: north coordinate of point
    :param transform: georeferencing transform of the image, e.g. `(x_upper_left, res_x, 0, y_upper_left, 0, -res_y)`
    :param truncate: Truncate pixel coordinates. Default is `True`
    :return: row and column pixel image coordinates
    """
    column = (east - transform[0]) / transform[1]
    row = (north - transform[3]) / transform[5]
    if truncate:
        return int(row + ERR), int(column + ERR)
    return row, column


def pixel_to_utm(row: float, column: float, transform: Sequence[float]) -> tuple[float, float]:
    """Convert pixel coordinate to UTM coordinate given a transform

    :param row: row pixel coordinate
    :param column: column pixel coordinate
    :param transform: georeferencing transform of the image, e.g. `(x_upper_left, res_x, 0, y_upper_left, 0, -res_y)`
    :return: east, north UTM coordinates
    """
    east = transform[0] + column * transform[1]
    north = transform[3] + row * transform[5]
    return east, north


@deprecated("The function `wgs84_to_pixel` has been deprecated.", category=SHDeprecationWarning)
def wgs84_to_pixel(
    lng: float, lat: float, transform: Sequence[float], utm_epsg: CRS | None = None, truncate: bool = True
) -> tuple[float, float] | tuple[int, int]:
    """Convert WGS84 coordinates to pixel image coordinates given transform and UTM CRS. If no CRS is given it will be
    calculated it automatically.

    :param lng: longitude of point
    :param lat: latitude of point
    :param transform: georeferencing transform of the image, e.g. `(x_upper_left, res_x, 0, y_upper_left, 0, -res_y)`
    :param utm_epsg: UTM coordinate reference system enum constants
    :param truncate: Truncate pixel coordinates. Default is `True`
    :return: row and column pixel image coordinates
    """
    east, north = wgs84_to_utm(lng, lat, utm_epsg)
    row, column = utm_to_pixel(east, north, transform, truncate=truncate)
    return row, column


def get_utm_crs(lng: float, lat: float, source_crs: CRS = CRS.WGS84) -> CRS:
    """Get CRS for UTM zone in which (lat, lng) is contained.

    :param lng: longitude
    :param lat: latitude
    :param source_crs: source CRS
    :return: CRS of the zone containing the lat,lon point
    """
    if source_crs is not CRS.WGS84:
        lng, lat = transform_point((lng, lat), source_crs, CRS.WGS84)
    return CRS.get_utm_from_wgs84(lng, lat)


def transform_point(
    point: tuple[float, float], source_crs: CRS, target_crs: CRS, always_xy: bool = True
) -> tuple[float, float]:
    """Maps point form src_crs to tgt_crs

    :param point: a tuple `(x, y)`
    :param source_crs: source CRS
    :param target_crs: target CRS
    :param always_xy: Parameter that is passed to `pyproj.Transformer` object and defines axis order for
        transformation. The default value `True` is in most cases the correct one.
    :return: point in target CRS
    """
    if source_crs == target_crs:
        return point
    transform_function = CRS.get_transform_function(source_crs, target_crs, always_xy=always_xy)
    return cast(Tuple[float, float], transform_function(*point))
